# -*- coding: utf-8 -*-
#
# badidatetime/_timedateutils.py
#
__docformat__ = "restructuredtext en"

import time
import locale
import math
from typing import NamedTuple

from ._structures import struct_time, ShortFormStruct, LongFormStruct
from .badi_calendar import BahaiCalendar


class TimeDateUtils(BahaiCalendar):
    """
    This class provides utility functionality to the datetime package. Its
    use is through a pre-instantiated object _td_utils.
    """
    # Badi additions are %:K for Kull-i-Shay and %:V for Váḥid.
    _VALID_FORMAT_CHRS = 'aAbBcCdDefGhHIjkKlmMnprSTuUVWxXyYzZ%'
    """
    str: A list of all the format characters.
    """
    DAYNAMES = ('Jalál', 'Jamál', 'Kamál', 'Fiḍāl', '`Idāl',
                'Istijlāl', 'Istiqlāl')
    """
    tuple: The full day names.
    """
    DAYNAMES_ABV = ('Jal', 'Jam', 'Kam', 'Fiḍ', 'Idā', 'Isj', 'Isq')
    """
    tuple: The abreviated day names.
    """
    MONTHNAMES = {1: 'Bahá', 2: 'Jalál', 3: 'Jamál', 4: "'Aẓamat", 5: 'Núr',
                  6: 'Raḥmat', 7: 'Kalimát', 8: 'Kamál', 9: "Asmá'",
                  10: "'Izzat", 11: 'Mashíyyat', 12: "'Ilm", 13: 'Qudrat',
                  14: 'Qawl', 15: 'Masá’il', 16: 'Sharaf', 17: 'Sulṭán',
                  18: 'Mulk', 0: 'Ayyám-i-Há', 19: "'Alá'"}
    """
    dict: The full month names.
    """
    MONTHNAMES_ABV = {1: 'Bah', 2: 'Jal', 3: 'Jam', 4: 'Aẓa', 5: 'Núr',
                      6: 'Raḥ', 7: 'Kal', 8: 'Kam', 9: 'Asm', 10: 'Izz',
                      11: 'Mas', 12: 'Ilm', 13: 'Qud', 14: 'Qaw', 15: 'Mas',
                      16: 'Sha', 17: 'Sul', 18: 'Mul', 0: 'Ayy', 19: 'Alá'}
    """
    dict: The abreviated month names.
    """
    FIRST_YEAR_EPOCH = 1721501.261143
    """
    float: The first day that we start our count, Julian year 1, March 19th.
    """
    DAYS_BEFORE_1ST_YEAR = 77
    """
    int: Keeps the Badí' day count in par with the Gregorian day count.
    """
    _SHORT_STRUCT_TM_ITEMS = 6
    """
    int: Length of the short form Badí' date and time portion of the
    struct time.
    """
    _LONG_STRUCT_TM_ITEMS = 8
    """
    int: Length of the long form Badí' date and time portion of the
    struct time.
    """

    def __init__(self):
        """
        We need to set the locale. However if an unsupported locale is
        required it may nee to be installed. This is the Linux method.

        1. Find the supported locales:
           $ locale -a
        2. Set a new locale:
           $ sudo locale-gen fr_FR.UTF-8 # Use the the locale you need.
           $ sudo update-locale
        """
        super().__init__(self)
        self._locale_data = {}
        self._date_and_time_locale()

    def _date_and_time_locale(self):
        """
        Set the `_locale_data` dictionary instance object to locale date
        and time information.
        """
        self._locale_data['locale'] = locale.setlocale(locale.LC_TIME, '')
        self._locale_data['am'] = locale.nl_langinfo(locale.AM_STR)
        self._locale_data['pm'] = locale.nl_langinfo(locale.PM_STR)

        try:
            # Get the date format for the current locale
            date_format = locale.nl_langinfo(locale.D_FMT)
        except AttributeError:  # pragma: no cover
            date_format = '%m/%d/%y'

        self._locale_data['d_format'] = self._order_format(
            date_format, '%m/%d/%y')
        self._locale_data['t_format'] = self._order_format(
            self._find_time_order(), '%H:%M:%S')

    def _order_format(self, fmt: str, default: str) -> None:
        """
        Pulls out of the incoming strings data needed to processing date
        and time formats.

        :param str fmt: The string format to parse.
        :param str default: A default format if `fmt` is invalid.
        :returns: Data for processing date and time formats.
        :rtype: str
        """
        if len(fmt) != 8:
            fmt = default

        data = [fmt[2]]

        for idx, char in enumerate(fmt):
            if idx in (1, 4, 7):
                data.append((char))

        return data

    def _find_time_order(self):
        """
        Find the locale time delimiter and id 24 or 12 hour time is used.

        :returns: The time delimiter and 24 or 12 hour time.
        :rtype: str
        """
        delim = time.strftime('%X')[2]
        h = 'I' if time.strftime('%p') != "" else 'H'
        return f"%{h}{delim}%M{delim}%S"

    @property
    def locale(self):  # pragma: no cover
        """
        Returns the current locale.
        """
        return self._locale_data['locale']

    @property
    def am(self):
        """
        Returns the AM designator.
        """
        return self._locale_data['am']

    @property
    def pm(self):
        """
        Returns the PM designator.
        """
        return self._locale_data['pm']

    @property
    def date_format(self):
        """
        Returns the locale's date format.
        """
        return self._locale_data['d_format']

    @property
    def time_format(self):
        """
        Returns the locale's time format.
        """
        return self._locale_data['t_format']

    def _checktm(self, ttup: tuple) -> None:
        """
        Check that the fields in the tuple are of the correct type. This
        check on date information is different than what is found inn the
        badi_calendar.py module as it needs to conform with ISO standards.

        :param ttup: A time tuple.
        :type ttup: ShortFormStruct or LongFormStruct
        """
        if not issubclass(ttup.__class__, tuple):
            raise TypeError(
                f"The ttup argument {ttup.__class__} is not a proper tuple.")

        def process_long_form(ttup: tuple):
            assert (self.KULL_I_SHAY_MIN <= ttup[0] <= self.KULL_I_SHAY_MAX), (
                f"Invalid kull-i-shay {ttup[0]}, it must be in the range of "
                f"[{self.KULL_I_SHAY_MIN}, {self.KULL_I_SHAY_MAX}].")
            assert 1 <= ttup[1] < 20, (
                f"Invalid Váḥids '{ttup[1]}' in a Kull-i-Shay’, it must be in "
                "the range of [1, 19].")
            assert 1 <= ttup[2] < 20, (
                f"Invalid year '{ttup[2]}' in a Váḥid, it must be in the "
                "range of [1, 19].")
            return (ttup[0] - 1) * 361 + (ttup[1] - 1) * 19 + ttup[2]

        def process_short_form(ttup: tuple):
            assert self.MINYEAR <= ttup[0] <= self.MAXYEAR, (
                f"Invalid year '{ttup[0]}' it must be in the range of ["
                f"{self.MINYEAR}, {self.MAXYEAR}].")
            return ttup[0]

        t_len = len(ttup)
        named_tuple = hasattr(ttup, '_asdict')  # _make also can work

        if named_tuple:  # Both long and short NamedTuple
            if t_len == 13:  # Long form
                year = process_long_form(ttup)
                idx = 3
            elif t_len == 11:  # Short form
                year = process_short_form(ttup)
                idx = 1
            else:  # pragma: no cover
                raise TypeError(f"Invalid timetuple, found length {t_len}.")
        else:  # A Tuple or class derived from a standard tuple
            if t_len == 11:  # Long form
                year = process_long_form(ttup)
                idx = 3
            elif t_len == 9:  # Short form
                year = process_short_form(ttup)
                idx = 1
            else:
                raise TypeError(f"Invalid timetuple, found length {t_len}.")

        month = ttup[idx]
        day = ttup[idx+1]
        hour = ttup[idx+2]
        minute = ttup[idx+3]
        second = ttup[idx+4]
        wday = ttup[idx+5]
        yday = ttup[idx+6]
        isdst = ttup[idx+7]
        assert 0 <= month < 20, (
            f"Invalid month '{month}', it must be in the range of [0, 19].")
        cycle = (4 + self._is_leap_year(year)) if month == 0 else 19
        assert 1 <= day <= cycle, (
            f"Invalid day '{day}' for month '{month}', it must be in the "
            f"range of [1, {cycle}].")
        assert 0 <= hour <= 24, (
            f"Invalid hour '{hour}', it must be in the range of [0, 24].")
        assert 0 <= minute < 60, (
            f"Invalid minute '{minute}', it must be in the range of [0, 60].")
        assert 0 <= second <= 61, (
            f"Invalid second '{second}', it must be in the range of [0, 61].")
        assert 0 <= wday <= 6, (
            f"Invalid week day '{wday}', it must be in the range of [0, 6].")
        assert 1 <= yday <= 366, (
            f"Invalid day '{yday}' in year, it must be in the range of "
            "[1, 366].")
        assert -1 <= isdst <= 1, (
            f"Invalid isdst '{isdst}', it must be in the range of [-1, 1].")

    def a(self, ttup, org, mod):
        """
        Abbreviated weekday.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The abbreviated weekday.
        :rtype: str
        """
        return self.DAYNAMES_ABV[ttup.tm_wday]

    def A(self, ttup, org, mod):
        """
        Full weekday name.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The full weekday.
        :rtype: str
        """
        return self.DAYNAMES[ttup.tm_wday]

    def b(self, ttup, org, mod):
        """
        Abbreviated month name.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The abbreviated month name.
        :rtype: str
        """
        return self.MONTHNAMES_ABV[ttup.tm_mon]

    def B(self, ttup, org, mod):
        """
        Full month name.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The full month name.
        :rtype: str
        """
        return self.MONTHNAMES[ttup.tm_mon]

    def c(self, ttup, org, mod):
        """
        Locale specific date and time. Equivalent to "%a %b %d %H:%M:%S %Y".

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The locale specific date and time.
        :rtype: str
        """
        st = f"{self.DAYNAMES_ABV[ttup.tm_wday]} "
        st += f"{self.MONTHNAMES_ABV[ttup.tm_mon]} "
        st += f"{ttup.tm_mday:02} "
        st += f"{ttup.tm_hour:02}:"
        st += f"{ttup.tm_min:02}:"
        st += f"{ttup.tm_sec:02} "

        if not ttup.short:
            st += f"{ttup.tm_kull_i_shay} "
            st += f"{ttup.tm_vahid:02} "
            st += f"{ttup.tm_year:02}"
        else:
            st += f"{ttup.tm_year:04}"

        return st

    def C(self, ttup, org, mod):
        """
        Century as a space-padded decimal number. (The year divided by 100
        then truncated to an integer.)

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The century as a space-padded decimal number.
        :rtype: str
        """
        year = self._get_year(ttup)
        n = '-' if year < 0 else ''
        return f"{n}{abs(math.floor(year / 100)):02}"

    def d(self, ttup, org, mod):
        """
        Day of month as a zero-padded decimal number.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The day of month as a zero-padded decimal number.
        :rtype: str
        """
        if mod == '-':
            st = f"{ttup.tm_mday}"
        else:
            pad = ' ' if org == 'e' else 0
            st = f"{ttup.tm_mday:{pad}2}"

        return st

    def D(self, ttup, org, mod):
        """
        Date where year is without century. Equivalent to a localized %m/%d/%y.

        .. note::

           Return a locale dependent Badí' short date. Badí' long dates are
           converted to short dates first.
           This method does not take into account format extenders, in other
           words the - or : after the %. They should never show up in the
           locale.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The date where year is without century.
        :rtype: str
        """
        year = self._get_year(ttup)
        century = int(year / 100) * 100
        year -= 0 if year < century else century
        sep = self.date_format[0]
        data = []

        for p in range(1, 4):
            fmt = self.date_format[p]

            if fmt[-1] in 'yY':
                data.append(f"{year:02}")
            else:
                org = fmt[0]
                mod = ''
                data.append(f"{getattr(self, org)(ttup, org, mod):02}")

        return sep.join(data)

    def f(self, ttup, org, mod):
        """
        Microseconds as a decimal number.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The Microseconds as a decimal number.
        :rtype: str
        """
        s, m = self._sec_microsec_from_seconds(ttup.tm_sec)
        return f"{round(m, 6):06}"

    def G(self, ttup, org, mod):
        """
        ISO 8601 year with century as a zero-padded decimal number.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The ISO 8601 year with century as a zero-padded decimal
                 number.
        :rtype: str
        """
        year = self._get_year(ttup)
        n = '-' if year < 0 else ''
        return f"{n}{abs(year):04}"

    def H(self, ttup, org, mod):
        """
        Hour (24-hour clock) as a decimal number. Either zero-padded if %H
        or not zero-padded if %-H.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The hour (24-hour clock) as a decimal number.
        :rtype: str
        """
        if mod == '-':  # %-H
            st = f"{ttup.tm_hour}"
        else:  # %H
            st = f"{ttup.tm_hour:02}"

        return st

    def I(self, ttup, org, mod):
        """
        Hour (12-hour clock) as a zero-padded decimal number.

        .. note::

           1. If we assume that sunset was at 1800 hrs UTC then the Badí' noon
              would be about 0600 hrs UTC the next morning. This changes on a
              daily bases because sunset changes and there is seldom exactly 24
              hours between two sunsets.
           2. 1st we need to find sunset for the provided date and the day
              after. Subtract these two times and divide the results by 2 to
              determine noon. Then determine which side of noon the current
              Badí' time is on.

        *** TODO *** Does a 12-hour clock make sense in a Badí' time?

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The hour (12-hour clock) as a zero-padded decimal number.
        :rtype: str
         """
        midday_frac = self._find_midday(ttup)
        time_frac = self._decimal_day_from_hms(ttup.tm_hour, ttup.tm_min,
                                               ttup.tm_sec)

        if midday_frac <= time_frac:
            hour = ttup.tm_hour - 12
        else:
            hour = ttup.tm_hour

        if org == 'l' and mod == '-':
            st = f"{hour}"
        elif org == 'l':
            st = f"{hour: 2}"
        else:  # %I
            st = f"{hour:02}"

        return st

    def j(self, ttup, org, mod):
        """
        Day of the year as a decimal number. Either zero-padded if %j or
        not zero-padded if %-j.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The day of the year as a decimal number.
        :rtype: str
        """
        return f"{ttup.tm_yday}" if mod == '-' else f"{ttup.tm_yday:03}"

    def K(self, ttup, org, mod):
        """
        Kull-i-Shay as a negative or positive decimal number.

        .. note::

           If the mod is not a : (colon) them return an empty string.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The Kull-i-Shay as a decimal number.
        :rtype: str
        """
        st = ""

        if mod == ':':
            if ttup.short:
                k = ttup.tm_year / 361
                kull_i_shay = 0 if ttup.tm_year == 0 else math.ceil(k)
            else:
                kull_i_shay = ttup.tm_kull_i_shay

            n = '-' if kull_i_shay < 0 else ''
            st += f"{n}{kull_i_shay}"

        return st

    def m(self, ttup, org, mod):
        """
        Month as a decimal number. Either zero-padded if %m or not
        zero-padded if %-m.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The month as a decimal number.
        :rtype: str
        """
        return f"{ttup.tm_mon}" if mod == '-' else f"{ttup.tm_mon:02}"

    def M(self, ttup, org, mod):
        """
        Minute as a zero-padded decimal number. Either zero-padded if %M or not
        zero-padded if %-M.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The minute as a zero-padded decimal number.
        :rtype: str
        """
        return f"{ttup.tm_min}" if mod == '-' else f"{ttup.tm_min:02}"

    def n(self, ttup, org, mod):
        """
        Newline character.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The newline character.
        :rtype: str
        """
        return "\n"

    def p(self, ttup, org, mod):
        """
        Locale defined AM and PM.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The locale defined AM and PM.
        :rtype: str
        """
        midday_frac = self._find_midday(ttup)
        time_frac = self._decimal_day_from_hms(ttup.tm_hour, ttup.tm_min,
                                               ttup.tm_sec)

        if midday_frac <= time_frac:
            st = self.pm
        else:
            st = self.am

        return st

    def r(self, ttup, org, mod):
        """
        Locale defined 12-hour clock time (am/pm).

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The locale defined 12-hour clock time.
        :rtype: str
        """
        sec = math.floor(ttup.tm_sec)

        if org == 'T':
            st = f"{ttup.tm_hour:02}:{ttup.tm_min:02}:{sec:02}"
        else:
            hour = self.I(ttup, '', '')
            p = self.p(ttup, '', '')
            st = hour + f":{ttup.tm_min:02}:{sec:02} " + p

        return st

    def S(self, ttup, org, mod):
        """
        Second as a decimal number. Either zero-padded if %S or not
        zero-padded if %-S.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The second as a decimal number.
        :rtype: str
        """
        return f"{ttup.tm_sec}" if mod == '-' else f"{ttup.tm_sec:02}"

    def u(self, ttup, org, mod):
        """
        Weekday as a decimal number. With Jalál as 1.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The weekday as a decimal number.
        :rtype: str
        """
        wday = ttup.tm_wday + (0 if org == 'w' else 1)
        return f"{wday}"

    def U(self, ttup, org, mod):
        r"""
        Week number of the year (Jalál as the first day of the week) as a
        zero-padded decimal number. All days in a new year preceding the
        first \`Idāl are considered to be in week 0.

        .. note::

           It makes no sense to start a week on different day in the Badí'
           Calendar. So the %W format is the same as %U.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The week number of the year.
        :rtype: str
        """
        year = self._get_year(ttup)
        year, week, day = self._year_week_day(year, ttup.tm_mon,
                                              ttup.tm_mday, week0=True)
        return f"{week:02}"

    def V(self, ttup, org, mod):
        """
        ISO 8601 week as a decimal number with Jalál as the first day of the
        week. Week 01 is the week containing the 4th of Bahá. Either
        zero-padded if %V or not zero-padded if %-V.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The ISO 8601 week as a decimal number.
        :rtype: str
        """
        if mod == ':':
            if ttup.short:
                date = self.long_date_from_short_date(
                    (ttup.tm_year, ttup.tm_mon, ttup.tm_mday))
                st = f"{date[1]:02}"
            else:
                st = f"{ttup.tm_vahid:02}"
        else:
            year = self._get_year(ttup)
            year, week, day = self._year_week_day(year, ttup.tm_mon,
                                                  ttup.tm_mday)
            st = f"{week:02}"

        return st

    def x(self, ttup, org, mod):
        """
        Locale defined date representation.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The locale defined date representation.
        :rtype: str
        """
        delim = self.date_format[0]
        data = []

        for fmt in self.date_format[1:]:
            data.append(getattr(self, fmt)(ttup, '', ''))

        return f"{delim}".join(data)

    def X(self, ttup, org, mod):
        """
        Locale defined time representation.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The locale defined time representation.
        :rtype: str
         """
        delim = self.time_format[0]
        data = []

        for fmt in self.time_format[1:]:
            data.append(getattr(self, fmt)(ttup, '', ''))

        return f"{delim}".join(data)

    def y(self, ttup, org, mod):
        """
        Year without century as a decimal number. Either zero-padded if %y
        or not zero-padded if %-y.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The year without century as a decimal number.
        :rtype: str
        """
        year = self._get_year(ttup)
        century = int(year / 100) * 100
        year = year - century
        return f"{year}" if mod == '-' else f"{year:02}"

    def Y(self, ttup, org, mod):
        """
        Year with century as a zero-padded decimal number.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The year with century as a decimal number.
        :rtype: str
        """
        year = self._get_year(ttup)
        n = '-' if year < 0 else ''
        return f"{n}{abs(year):04}"

    def z(self, ttup, org, mod):
        """
        UTC offset in the form ±HHMM[SS[.ffffff]] (empty string if the
        object is naive).

        .. note::

           Some equivalents from offset to ISO standard.
           -14400.0       == -0400
           37080          == +1030
           22829.4        == +063415
           11056.44427776 == +030712.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The UTC offset in the form ±HHMM[SS[.ffffff]].
        :rtype: str
        """
        st = ""

        if ttup.tm_gmtoff:
            n = '-' if ttup.tm_gmtoff < 0 else '+'
            h = abs(ttup.tm_gmtoff / 3600)
            hh = int(h)
            m = h - hh
            m0 = m * 100
            mm = math.floor(m0)
            s = m0 - mm
            ss = int(s * 100)
            ms = int((s * 100 - ss) * 1000000)
            delim = ':' if mod == ':' else ""
            st += f"{n}{hh:02}{delim}{mm:02}"
            st += f"{delim}{ss:02}" if ss > 0 else ""
            st += f".{ms:06}" if ms > 0 else ""

        return st

    def Z(self, ttup, org, mod):
        """
        Time zone name (empty string if the object is naive).

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The time zone name.
        :rtype: str
        """
        return f"{ttup.tm_zone}" if ttup.tm_zone else ""

    def percent(self, ttup, org, mod):
        """
        The literal '%' character.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param str, org: The original text code.
        :param str mod: Code modifier.
        :returns: The literal '%' character.
        :rtype: str
        """
        return "%"

    __METHOD_LOOKUP = {'a': a, 'A': A, 'b': b, 'B': B, 'c': c, 'C': C, 'd': d,
                       'D': D, 'e': d, 'f': f, 'G': G, 'h': b, 'H': H, 'I': I,
                       'j': j, 'k': H, 'K': K, 'l': I, 'm': m, 'M': M, 'm': m,
                       'M': M, 'n': n, 'p': p, 'r': r, 'S': S, 'T': r, 'u': u,
                       'U': U, 'V': V, 'w': u, 'W': U, 'x': x, 'X': X, 'y': y,
                       'Y': Y, 'z': z, 'Z': Z, '%': percent
                       }
    """
    dict: An internal list of the format methods.
    """

    def strftime(self, format: str, ttup: tuple, tzinfo=None) -> str:
        """
        Convert a struct_time object into a string according to a specified
        format.

        :param str format: A string format.
        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :param tzinfo tzinfo: A tzinfo object.
        :returns: A struck_time object converted to a formatted string.
        :rtype: str
        """
        self._check_format(format)
        self._checktm(ttup)

        if not isinstance(ttup, (ShortFormStruct, LongFormStruct)):
            ttup = struct_time(ttup, tzinfo=tzinfo)

        idx, fmtlen = 0, len(format)
        strf = ""

        while idx < fmtlen:
            ch = format[idx]

            if ch == '%' and idx+1 < fmtlen:
                ch0 = format[idx+1]
                i = 2 if ch0 in '-:' else 1
                ch1 = format[idx+i]
                strf += self.__METHOD_LOOKUP[ch1](
                    self, ttup, ch1, ch0 if i == 2 else '')
            elif format[idx-1] not in '%-:':
                strf += ch

            idx += 1

        return strf

    def _check_format(self, format: str) -> None:
        """
        Check that the correct format was provided.

        :param str format: A string format.
        """
        idx = 0
        fmtlen = len(format)

        while idx < fmtlen:
            ch = format[idx]

            if ch == '%' and format[idx-1] != '%':
                ch0 = format[idx+1]
                i = 2 if ch0 in '-:' else 1
                ch1 = format[idx+i]

                if ((ch1 not in self._VALID_FORMAT_CHRS) or
                    (ch0 == '-' and ch1 not in 'dHjlmMSy') or
                    (ch0 == ':' and ch1 not in 'KVz')):
                    raise ValueError(
                        f"Invalid format character '{format[idx:idx+i+1]}'")

            idx += 1

        if fmtlen == 0:
            raise ValueError("Found an empty format string.")

    def _find_midday(self, ttup: tuple) -> tuple:
        """
        Midday time in hours, minutes, and seconds representing the Badí'
        midday.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :returns: The Badí' midday.
        :rtype: tuple
        """
        if ttup.short:
            date = (ttup.tm_year, ttup.tm_mon, ttup.tm_mday, ttup.tm_hour,
                    ttup.tm_min, ttup.tm_sec)
        else:
            date = (ttup.tm_kull_i_shay, ttup.tm_vahid, ttup.tm_year,
                    ttup.tm_mon, ttup.tm_mday, ttup.tm_hour, ttup.tm_min,
                    ttup.tm_sec)

        return self.midday(date)

    def _get_year(self, ttup: tuple) -> int:
        """
        Convert The Kull-i-Shay', Váḥid, year to a short mode year.

        :param ttup: A struct_time object.
        :type ttup: ShortFormStruct or LongFormStruct
        :returns: The short form year converted if necessary.
        :rtype: int
        """
        return (ttup.tm_year if ttup.short else
                ((ttup.tm_kull_i_shay - 1) * 361 + (ttup.tm_vahid - 1) * 19 +
                 ttup.tm_year))

    def _year_week_day(self, year: int, month: int, day: int,
                       week0: bool=False) -> tuple:
        """
        Return the year, week, and day of the week from a short form
        Badí' date.

        :param int, year: The year.
        :param int month: The month.
        :param int day: The day.
        :returns: The year, week, and day of the week.
        :rtype: tuple
        """
        week1jalal = self._isoweek1jalal(year)
        today = self._ymd2ord(year, month, day)
        # Internally, week and day have origin 0
        week, day = divmod(today - week1jalal, 7)

        if not week0:
            if week < 0:
                year -= 1
                week1jalal = self._isoweek1jalal(year)
                week, day = divmod(today - week1jalal, 7)
            elif week >= 52:
                if today >= self._isoweek1jalal(year+1):
                    year += 1
                    week = 0

        return year, week+1, day+1

    def _days_before_year(self, year: int) -> float:
        """
        Get the number of days before the 1st of Bahá of the year.

        :param int year: Badí' year
        :returns: The number of days since (-1841, 19, 19) of the Badí'
                  calendar.
        :rtype: int
        """
        jd0 = self.jd_from_badi_date((self.MINYEAR-1, 19, 19), _chk_on=False)
        jd1 = self.jd_from_badi_date((year, 1, 1), _chk_on=False)
        return math.floor(jd1 - jd0) - 1

    def _days_in_month(self, year: int, month: int) -> int:
        """
        The number of days in provided month in provided year.

        :param int year: Badí' year
        :param int month: Badí' month (0..19)
        :returns: The number of in the current month.
        :rtype: int
        """
        return 4 + self._is_leap_year(year) if month == 0 else 19

    def _days_before_month(self, year: int, month: int) -> int:
        """
        The number of days in the year preceding the first day of month.

        :param int year: Badí' year
        :param int month: Badí' month (0..19)
        :returns: The number in the year preceding the first day of month.
        :rtype: int
        """
        month -= -18 if month < 2 else 1 if 1 < month < 19 else 19
        dbm = 0

        if 0 < month < 19:
            dbm += month * 19
        elif month == 0:
            dbm += 18 * 19 + 4 + self._is_leap_year(year)

        return dbm

    def _day_of_week(self, year: int, month: int, day: int) -> int:
        """
        Find the day of the week where 0 == Jalál (Saturday) and
        6 == Istiqlāl (Friday).

        :param int year: Badí' year
        :param int month: Badí' month (0..19)
        :param int day: Badí' day
        :returns: The numerical day of the week.
        :rtype: int
        """
        # Since the usual start day is Monday (Kamál) a properly aligned
        # day number to the day name we need to add 1 to the ordinal.
        return ((self._ymd2ord(year, month, day) + 1) % 7 + 7) % 7

    def _ymd2ord(self, year: int, month: int, day: int) -> int:
        """
        Get the number of days since Badí' year -1842 (Julian 0001-03-19)
        including the current day.

        year, month, day -> ordinal, considering -1842-01-01 as day 1

        :param int year: Badí' year
        :param int month: Badí' month [0, 19]
        :param int day: Badí' day
        :returns: The number of days since Badí' year -1842 including the
                  current day.
        :rtype: int
        """
        dim = self._days_in_month(year, month)
        assert 1 <= day <= dim, (
            f"Day '{day}' for month {month} must be in range of 1..{dim}")
        # We add 77 days to the total so that the ordinal number can be
        # compared to the ordinals in the standard datetime package.

        # For some reason out of the 3004 year that are provided only
        # these three years are off by 1.
        if year in (-1796, -1792, -1788):
            fudge = 1
        else:
            fudge = 0

        return (self.DAYS_BEFORE_1ST_YEAR + self._days_before_year(year) +
                self._days_before_month(year, month) + day + fudge)

    def _ord2ymd(self, n: int, *, short: bool=False) -> tuple:
        """
        It is difficult to do this in the Badí' Calendar because a Badí' day
        can be more or less than 24 hours depending on when sunset is and the
        time of the year. From the summer Solstice to the winter Solstice the
        days get shorter. The day slowly comes down to 24 hours around the
        Fall Equinox and then below 24 hours. The inverse happens between the
        Winter Solstice and the Summer Solstice. We just use the BadiCalendar
        API.

        :param int n: The ordinal number of days from the MINYEAR.
        :param bool short: If True then parse for a short date else if False
                           parse for a long date.
        :returns: The Badí' date.
        :rtype: tuple
        """
        # We subtract 77 days from the total then add the value of n so that
        # the Badí' date will be the same as the date value passed into
        # _ymd2ord and give the same date as Python standard datetime package.
        # The reason we need to do this is that the first date that this
        # package can provide is equivalent to Julian year 1, March, 19th.
        jd = self.FIRST_YEAR_EPOCH - 1 - self.DAYS_BEFORE_1ST_YEAR + n
        return self.badi_date_from_jd(math.floor(jd) + 0.5, short=short,
                                      trim=True, rtd=True, _chk_on=False)

    def _build_struct_time(self, date: tuple, dstflag: int, *, tzinfo=None,
                           short_in: bool=False) -> NamedTuple:
        """
        Build either a ShortFormStruct or LongFormStruct struct_time.

        :param tuple date: A tuple date and time object.
        :param int dstflag: A flag indicating daylight savings time. May be
                            set to 1 when daylight savings time is in effect,
                            and 0 when it is not. A value of -1 indicates
                            that this is not known.
        :param tzinfo tzinfo: If provided a tzinfo object.
        :returns: A struct_time object.
        :rtype: ShortFormStruct or LongFormStruct
        """
        if short_in:
            y, m, d, hh, mm, ss = date
        else:
            # Microsecond (ms) is not used.
            y, m, d, hh, mm, ss, ms = self.short_date_from_long_date(
                date, _chk_on=False)

        wday = self._day_of_week(y, m, d)
        dnum = self._days_before_month(y, m) + d
        return struct_time(date + (wday, dnum, dstflag), tzinfo=tzinfo)

    def _isoweek_to_badi(self, year: int, week: int, day: int, *,
                         short: bool=False) -> tuple:
        """
        The week counts from Jalál (Saturday) as the first day and Istiqlāl
        (Friday) the last day of the week. This is different from the usual
        way ISO weeks are counted in the Gregorian Calendar which is Monday
        to Sunday.

        .. note::

           Whereas a Gregorian year can have 53 weeks in it a Badí' year can
           have 51 weeks in it and never 53.

        :param int year: Badí' year.
        :param int month: Badí' month (0..19)
        :param int day: Badí' day in week.
        :param bool short: If True then parse for a short date else if False
                           parse for a long date.
        :returns: A Badí' date.
        :rtype: tuple
        :raises AssertionError: If the week or weekday is out of range.
        """
        p_offset = 0

        if not 0 < week < 52:  # We're looking for only the 53rd week.
            day_one = _td_utils._day_of_week(year, 1, 1) + 1

            if day_one in (3, 4):
                out_of_range = True

                if week == 52:
                    # In Badí' years that have 52 weeks and start on the 3rd
                    # day (Kamál) or the 4th day (Fiḍāl).
                    # Badí' weeks start on Jalal (Saturday).
                    # *** TODO *** This needs to be tested over a larger year
                    #              spread.
                    p_offset = 7
                    out_of_range = False

                assert not out_of_range, f"Invalid week: {week}"

        assert 0 < day < 8, f"Invalid weekday: {day} (range is [1, 7])"
        # Now compute the offset from (Y, 1, 1) in days:
        day_offset = (week - 1) * 7 + (day - 1) + p_offset
        # Calculate the ordinal day for Jalal, week 1
        day_1 = self._isoweek1jalal(year)
        ord_day = day_1 + day_offset
        return self._ord2ymd(ord_day, short=short)

    def _isoweek1jalal(self, year: int) -> int:
        """
        Calculate the day number of Jalál (Saturday) starting week 1. It
        would be the first week with 4 or more days in the year in question.

        :param int year: Badí' year
        :returns: The number of the first Jalál in the Badí' year.
        :rtype: int
        """
        firstday = self._ymd2ord(year, 1, 1)
        # We subtract 6 instead of add 6 as is done in _isoweek1Monday.
        firstweekday = (firstday - 6) % 7
        week1jalal = firstday - firstweekday

        if firstweekday > 3:  # First week day >= Fidal
            week1jalal += 7

        return week1jalal

    def _parse_isoformat_date_time_timezone(self, dtstr: str) -> tuple:
        """
        Parse both the date and time represented in an ISO string into a
        date and time tuple.

        :param str dtstr: A ISO compliant time string.
        :returns: The date, time, and timezone.
        :rtype: tuple, tuple, timezone
        """
        def find_index(string, lst):
            indexes = [i for c in lst if (i := string.rfind(c)) > -1]
            return indexes[0] if len(indexes) > 0 else len(string)

        tz_chars = ('Z', 'B', '+', '-')
        init_chars = ('T', ' ') + tz_chars
        idx = find_index(dtstr, init_chars)
        str_date = dtstr[:idx]
        date = self._parse_isoformat_date(str_date) if str_date else ()
        str_other = dtstr[idx:]
        tz_1st = str_other[0] in tz_chars

        if tz_1st:
            tz = (self._parse_isoformat_timezone(str_other)
                  if str_other else None)
            time = ()
        else:
            idx = find_index(str_other, tz_chars)
            str_time = str_other[:idx]
            str_tz = str_other[idx:]
            time = self._parse_isoformat_time(str_time) if str_time else ()
            tz = self._parse_isoformat_timezone(str_tz) if str_tz else None

        return date, time, tz

    def _parse_isoformat_date(self, dtstr: str) -> tuple:
        """
        Parse a date ISO formatted string.

        :param str dtstr: A ISO compliant time string.
        :returns: The year, month, and day parsed from a ISO string.
        :rtype: tuple
        :raises AssertionError: Raised when the year is out of range or when
                                too many hyphens are used.
        :raises IndexError: When a string index is out of range.
        :raises ValueError: Raised when an invalid string is being parsed to
                            an integer or when an invalid ISO string is being
                            parsed.
        """
        for c in filter(lambda x: not x.isnumeric(), dtstr):
            if c not in ('-', 'W'):
                raise ValueError(
                    f"Invalid character {c!r} in incoming date string.")

        if dtstr != '':
            neg = dtstr[0] == '-'
            year = int(dtstr[:4 + neg])
            assert _td_utils.MINYEAR <= year <= _td_utils.MAXYEAR, (
                f"Year is out of range: {year}, min {_td_utils.MINYEAR}, "
                f"max {_td_utils.MAXYEAR}.")
            dtstr = dtstr[1:] if neg else dtstr

        dc = dtstr.count('-')
        wc = dtstr.count('W')
        assert ((wc == 0 and dc in (0, 1, 2)) or
                (wc == 1 and dc in (0, 1, 2))), (
                "Invalid format, there must be between 0 to 2 hyphens (-) in "
                "the date format or there can be one uppercase (W) week "
                "identifier and between 0 and 2 hyphens (-) used.")
        d_len = len(dtstr)

        if dc == 1 and d_len == 7 and not wc:    # YYYY-MM
            date = (year, int(dtstr[5:7]), 1)
        elif dc == 0 and d_len == 8 and not wc:  # YYYYMMDD
            date = (year, int(dtstr[4:6]), int(dtstr[7:9]))
        elif dc == 2 and not wc:                 # YYYY-MM-DD
            date = (year, int(dtstr[5:7]), int(dtstr[8:10]))
        # YYYYWww, YYYY-Www, YYYYWwwD, YYYY-Www-D
        elif wc and 7 <= d_len <= 10:
            pos = 5 if dc == 0 else 6
            wday = int(dtstr[pos:pos+2])
            pos += 2 if dc == 0 else 3
            d = dtstr[pos:]
            assert (dc == 1 and d_len == 8) or dc in (0, 2), (
                f"Invalid ISO string {dtstr}.")
            day = int(d) if d.isdigit() else 1
            date = self._isoweek_to_badi(year, wday, day, short=True)[:3]
        elif d_len in (7, 8):                    # YYYYDDD or YYYY-DDD
            month_days = self._BADI_MONTH_NUM_DAYS
            month_days[18] = (0, 4 + self._is_leap_year(year))
            days = int(dtstr[4:7] if dc == 0 else dtstr[5:8])

            for month, ds in month_days:
                if days <= ds: break
                days -= ds

            date = (year, month, days)
        else:
            date = ()

        return date

    def _parse_isoformat_time(self, tmstr: str) -> tuple:
        """
        Parse a time ISO formatted string.

        :param str tmstr: A ISO compliant time string.
        :returns: The hour, minute, and second parsed from an ISO string.
        :rtype: tuple
        :raises AssertionError: Raised when there are invalid time designators,
                                when to many colons used, or when too many dots
                                are used.
        :raises ValueError: Raised when an invalid string is being parsed to
                            an integer or when an invalid ISO string is being
                            parsed.
        """
        for c in filter(lambda x: not x.isnumeric(), tmstr):
            if c not in ('T', ' ', ':', '.'):
                raise ValueError(
                    f"Invalid character {c!r} in incoming time string.")

        t_len = len(tmstr)
        tmp_tmstr = tmstr
        tc = tmstr.count('T')
        sc = tmstr.count(' ')
        assert ((tc == 0 and sc == 0) or
                (tc == 1 or sc == 1) and (tc or sc) and not (tc and sc)), (
            "Cannot have both a 'T' and a space or more than one of either to "
            "indicate time.")

        if sc:
            tmstr = "T" + tmstr[1:]
            del sc
            tc = 1

        if t_len > 0 and 'T' != tmstr[0]:
            raise ValueError("Invalid time string, 1st character must be "
                             f"one of ( T), found {tmp_tmstr!r}")

        del tmp_tmstr
        cc = tmstr.count(':')
        assert cc < 3, (
            f"Invalid number of colons (:), can be 0 - 2, found {cc}")
        pc = tmstr.count('.')
        assert pc <= 1, f"Invalid number of dots (.), can be 0 - 1, found {pc}"

        if t_len > 2:
            hour = int(tmstr[1:3])
            pos0 = 1 if cc else 0
            pos1 = 2 if cc == 2 else 0

            if t_len > 3:
                if tmstr[3] == '.':  # Thh.hhh
                    ph = float(tmstr[3:]) * 60
                    minute = math.floor(ph)
                    second = (ph % 1) * 60
                    second = math.floor(second) if second % 1 == 0 else second
                    time = (hour, minute, second)
                elif tmstr[5 + pos0:6 + pos0] == '.':
                    # Thhmm.mmm or Thh:mm.mmm
                    minute = int(tmstr[3 + pos0:5 + pos0])
                    pm = float(tmstr[5 + pos0:])
                    second = pm * 60
                    second = math.floor(second) if second % 1 == 0 else second
                    time = (hour, minute, second)
                elif tmstr[7 + pos0:8 + pos0] == '.':
                    # Thhmmss.sss or Thh:mm:ss.sss
                    minute = int(tmstr[3 + pos0:5 + pos0])
                    second = float(tmstr[5 + pos0:])
                    second = math.floor(second) if second % 1 == 0 else second
                    time = (hour, minute, second)
                elif t_len == 5 + pos0:  # Thhmm or Thh:mm
                    minute = int(tmstr[3 + pos0:5 + pos0])
                    time = (hour, minute, 0)
                elif t_len >= 7 + pos1:  # Thhmmss.sss or Thh:mm:ss.sss
                    minute = int(tmstr[3 + pos0:5 + pos0])
                    second = float(tmstr[5 + pos1:])
                    second = math.floor(second) if second % 1 == 0 else second
                    time = (hour, minute, second)
                else:
                    raise ValueError(f"Invalid time string, found {tmstr!r}")
            else:  # Thh
                time = (hour, 0, 0)
        else:
            time = ()

        return time

    def _parse_isoformat_timezone(self, tzstr: str) -> tuple:
        """
        Parse a timezone ISO formatted string.

        :param str tzstr: A ISO compliant time string.
        :returns: A timezone object indicating the offset from UTC.
        :rtype: timezone
        :raises AssertionError: Raised when there are invalid timezone
                                delimiters are found.
        :raises ValueError: Raised when an invalid ISO string is being parsed.
        """
        from .datetime import timezone, timedelta

        for c in filter(lambda x: not x.isnumeric(), tzstr):
            if c not in ('Z', 'B', '+', '-', ':'):
                raise ValueError(
                    f"Invalid character {c!r} in incoming timezone string.")

        tz_len = len(tzstr)
        tmp_tzstr = tzstr
        nc = tzstr.count('-')
        pc = tzstr.count('+')
        zc = tzstr.count('Z')
        bc = tzstr.count('B')  # This is an extension to the ISO standard
        c_none = all([True for c in (nc, pc, zc, bc) if c == 0])  # All == 0
        ct_gt_1 = sum((nc, pc, zc, bc)) > 1  # More than one == 1
        ca_gt_1 = any([True for c in (nc, pc, zc, bc) if c > 1])  # Any > 1
        assert c_none and not ct_gt_1 and not ca_gt_1, (
            "Can only have one of (-+Z) and no more than one of (-+Z) to "
            "indicate a timezone.")

        if tz_len > 0 and ('-' != tzstr[0] and '+' != tzstr[0] and
                           'Z' != tzstr[0]) and 'B' != tzstr[0]:
            raise ValueError("Invalid timezone string, 1st character must be "
                             f"one of (-+Z), found {tmp_tzstr!r}")

        del tmp_tzstr
        cc = tzstr.count(':')
        assert cc < 2, (
            f"Invalid number of colons (:), can be 0 - 1, found {cc}")

        if tz_len > 0:
            if zc == 1 and tz_len == 1:
                tz = timezone.utc
            elif bc == 1 and tz_len == 1:
                tz = timezone.badi
            else:
                offset = [int(x) for x in tzstr[1:].split(':')]
                offset[0] = offset[0] * -1 if nc else offset[0]
                offset += [0] if len(offset) == 1 else []
                tz = timezone(timedelta(hours=offset[0], minutes=offset[1]))
        else:
            tz = None

        return tz

    def _check_date_fields(self, a: int, b: int, c: int, d: int=None,
                           e: int=None, *, short_in: bool=False) -> None:
        """
        Check the validity of the date.

        :param int a: The long form Kull-i-Shay or short form year.
        :param int b: The long form Váḥid or short form month.
        :param int c: The long form year or short form day.
        :param int d: The long form month.
        :param int e: The long form day.
        :param bool short_in: If True then parse for a short date else if False
                              parse for a long date. This is for incoming dates
                              not outgoing dates as in most other uses of
                              'short'.
        :returns: Nothing
        :rtype: None
        :raises AssertionError: If any of the date values are out of range.
        """
        if short_in:
            b_date = (a, b, c)
        else:
            b_date = (a, b, c, d, e)

        self._check_valid_badi_date(b_date, short_in=short_in)

    def _check_time_fields(self, hour: int, minute: int, second: int,
                           microsecond: int, fold: int) -> None:
        """
        Check the validity of the time.

        :param int hour: The hour.
        :param int minute: The minute.
        :param int second: The second.
        :param int microsecond: The microsecond.
        :param int fold: The value of 1 if the time is in the time fold when
                         the time falls back one hour in the Autumn or 0 any
                         other time of the year.
        """
        self._check_valid_badi_time(hour, minute, second, microsecond,
                                    maxsec=61)
        assert fold in (0, 1), (
            f"The fold argument '{fold}' must be either 0 or 1.")

    def _wrap_strftime(self, object, format: str,
                       timetuple: tuple, tzinfo=None) -> str:
        """
        Correctly substitute for %z and %Z escapes in strftime formats.

        :param class object: A class instance that uses this method.
        :param str format: The formatted string.
        :param timetuple:
        :type ttup: ShortFormStruct or LongFormStruct
        :returns: A correctly formatted string.
        :rtype: str
        """
        # Don't call utcoffset() or tzname() unless actually needed.
        freplace = None  # the string to use for %f

        # Scan format for %z and %Z escapes, replacing as needed.
        newformat = []
        push = newformat.append
        i, n = 0, len(format)
        tzinfo = object.tzinfo if hasattr(object, 'tzinfo') else None

        while i < n:
            ch = format[i]
            i += 1

            if ch == '%':
                if i < n:
                    ch = format[i]
                    i += 1

                    if ch == 'f':
                        if freplace is None:
                            freplace = f'{getattr(object, "microsecond", 0):06}'

                        push(freplace)
                    else:
                        push('%')
                        push(ch)
                else:
                    push('%')
            else:
                push(ch)

        newformat = "".join(newformat)
        return self.strftime(newformat, timetuple, tzinfo=tzinfo)


_td_utils = TimeDateUtils()



# -*- coding: utf-8 -*-
#
# badidatetime/gregorian_calendar.py
#
__docformat__ = "restructuredtext en"

import math

from badidatetime.base_calendar import BaseCalendar


class GregorianCalendar(BaseCalendar):
    """
    Implementation of the Gregorian Calendar.
    """
    # Julian date for the Gregorian epoch
    # https://www.grc.nasa.gov/www/k-12/Numbers/Math/Mathematical_Thinking/calendar_calculations.htm
    _GREGORIAN_EPOCH = 1721423.5
    _MONTHS = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
    _GREGORIAN_LEAP_YEAR = lambda self, year: (
        (year % 4 == 0) * ((year % 100 != 0) + (year % 400 == 0)) == 1)
    _GREGORIAN_LEAP_YEAR_ALT = lambda self, year: (
        (year % 4 == 0) * (year % 128 != 0) == 1)

    def __init__(self) -> None:
        super().__init__()
        # [year, month, day]
        self._gregorian_date = None

    def jd_from_gregorian_date(self, g_date: tuple, *, exact: bool=False,
                               alt: bool=False) -> float:
        """
        Convert Gregorian dates to Julian day count with the 1582 10, 15
        correction.

        :param tuple g_date: A Gregorian date in the (year, month, day) format.
        :param bool exact: Julian days as if the Gregorian calendar started on
                           year 1. This is astronomically correct but not
                           historically correct.
        :param bool alt: Use a more accurate leap year calculation, only valid
                         when the `exact` keyword is used, there is no effect
                         otherwise.
        :returns: A Julian day.
        :rtype: float

        .. note::

           1. See Astronomical Formulae for Calculators Enlarged & Revised,
              by Jean Meeus ch3 p24-25
           2. See: https://www.fourmilab.ch/documents/calendar/
              https://core2.gsfc.nasa.gov/time/julian.html
           3. Caution when using the `exact=True` keyword the Julian Period
              days returned will not always be the same when `exact=False`
              is used. This means date comparisons will be in error if both
              dates do not use the same True or False `exact` keyword.
        """
        year, month, day = self.date_from_ymdhms(g_date)

        if exact:  # An astronomically correct algorithm
            GLY = (self._GREGORIAN_LEAP_YEAR_ALT if alt
                   else self._GREGORIAN_LEAP_YEAR)
            td = self._days_in_years(year-1, alt=alt)
            days = td + (self._GREGORIAN_EPOCH - 1)
            month_days = list(self._MONTHS)
            month_days[1] += GLY(year)
            days += sum(month_days[:month-1]) + day
            jd = round(days, self._ROUNDING_PLACES)
        else:  # Meeus historically correct algorithm
            if (year, month) == (1582, 10):
                assert day not in (5, 6, 7, 8, 9, 10, 11, 12, 13, 14), (
                    f"The days 5-14 in 1582-10 are invalid, found day '{day}'."
                    )

            if month <= 2:
                year -= 1
                month += 12

            if (year, month, day) >= (1582, 10, 15):
                a = math.floor(year / 100)
                b = 2 - a + math.floor(a / 4)
            else:
                b = 0

            jd = round(math.floor(self._JULIAN_YEAR * year) + math.floor(
                30.6001 * (month + 1)) + day + b + 1720994.5,
                       self._ROUNDING_PLACES)

        return jd

    def gregorian_date_from_jd(self, jd: float, *, exact: bool=False,
                               alt: bool=False) -> tuple:
        """
        Convert Julian day to Gregorian date.

        :param float jd: A Julian period day. This value should be an
                         historical JD if exact is False and an astronomical
                         JD if exact is True.
        :param bool exact: If `False` (default) Meeus' historically algorithm
                           is used, else if `True` the more astronomically
                           correct algorithm is used.
        :param bool alt: If `False` (default) the more common 4|100|400
                         algorithm for leap years is used else if `True` the
                         less common 4|128 algorithm for leap years is used.
                         This argument only has an effect if the `exact`
                         argument is also used.
        :returns: A Gregorian date in the (year, month, day) format.
        :rtype: tuple

        .. note::

           See Astronomical Formulae for Calculators Enlarged & Revised,
           by Jean Meeus ch3 p26-29
        """
        if exact:  # An astronomically correct algorithm.
            GLY = (self._GREGORIAN_LEAP_YEAR_ALT if alt
                   else self._GREGORIAN_LEAP_YEAR)
            # Get the number of days since the Gregorian epoch.
            md = jd - (self._GREGORIAN_EPOCH - 1)
            year = math.floor(abs(md / self._MEAN_TROPICAL_YEAR)) + 1
            year *= -1 if md < (self._GREGORIAN_EPOCH - 1) else 1
            # Refine the number of days since the epoch for the date.
            td = self._days_in_years(year, alt=alt)
            days = md - td

            while days > 365:
                year += 1
                td = self._days_in_years(year, alt=alt)
                days = md - td

            if days == 0:
                days = 365 if year < 0 and year % 4 != 0 else 366
            else:
                year += 1

            month_days = list(self._MONTHS)
            month_days[1] += GLY(year)
            d = day = 0

            for month, ds in enumerate(month_days, start=1):
                d += ds
                if days > d: continue
                day = math.ceil(days - (d - ds))
                break

            f = jd % 1
            day += f - (1.5 if f > 0.5 else 0.5)

            if day < 1:
                month -= 1 if month > 1 else -11
                day = month_days[month-1] + day
                year -= 1 if month == 12 else 0

            date = (year, month, round(day, self._ROUNDING_PLACES))
        else:  # Meeus' historically correct algorithm.
            j_day = jd + 0.5
            z = math.floor(j_day)
            f = j_day % 1

            if z >= 2299161:  # 1582-10-15 Julian and Gregorian crossover.
                alpha = math.floor((z - 1867216.25) / 36524.25)
                a = z + 1 + alpha - math.floor(alpha / 4)
            else:
                a = z

            b = a + 1524
            c = math.floor((b - 122.1) / 365.25)
            d = math.floor(365.25 * c)
            e = math.floor((b - d) / 30.6001)
            day = b - d - math.floor(30.6001 * e) + f
            month = 0
            year = 0

            if e > 13:
                month = e - 13
            else:
                month = e - 1

            if month > 2:
                year = c - 4716
            else:
                year = c - 4715

            date = (year, month, round(day, self._ROUNDING_PLACES))

        return date

    def posix_timestamp(self, t: float, *, zone: float=0,
                        us: bool=False) -> tuple:
        """
        Find the year, month, day, hours, minutes, and seconds from a
        POSIX timestamp updated for provided time zone.

        .. warning::

           DO NOT USE THIS METHOD -- It's experimental and is not used in any
           other code.

        *** TODO *** Needs to be fixed for years before 1970 Gregorian.

        :param float t: POSIX timestamp
        :param float zone: Timezone
        :param bool us: If True the seconds are split to seconds amd
                        microseconds else if False the seconds has a
                        fractional day as a decimal.
        :returns: The time of the day corrected for the timezone.
        :rtype: tuple
        """
        t += zone * 3600
        days = math.floor(t / 86400)
        year = 1970
        leap = False
        days = abs(days) + 1

        # Find the year and remaining number of days.
        while True:
            leap = self._GREGORIAN_LEAP_YEAR(year)
            diy = 365 + leap
            if days < diy: break
            days -= diy
            year += 1

        month_days = list(self._MONTHS)
        month_days[1] += leap
        days_before_month = 0

        for month, days_in_month in enumerate(month_days, start=1):
            days_before_month += days_in_month
            if days > days_before_month: continue
            day = days - (days_before_month - days_in_month)

            # if day > days_in_month:
            #     month += 1
            #     day = day - days_in_month

            break

        seconds = t % 86400
        minutes = math.floor(seconds / 60)
        minute = minutes % 60
        hour = math.floor(minutes / 60)
        second = round(seconds % 60, 6)
        microsecond = 0

        if us:
            microsecond = self._PARTIAL_SECOND_TO_MICROSECOND(second)
            second = math.floor(second)

        # print(f"{t:18.6f} {zone:>+2.2f} {year:02} {month:02} {day:02} "
        #      f"{hour:02} {int(minute):02} {second:02} {microsecond}")

        date = (year, month, day, hour, minute, second
                ) + ((microsecond,) if us else ())
        self._check_valid_gregorian_month_day(date, historical=True)
        return date

    def gregorian_year_from_jd(self, jd: float) -> int:
        """
        Find the Gregorian year from a Julian Period day.

        :param float jd: The Julian Period day.
        :returns: The year portion of the Julian day.
        :rtype: int
        """
        return self.gregorian_date_from_jd(jd)[0]

    def date_from_ymdhms(self, date: tuple) -> tuple:
        """
        Convert (year, month, day, hour, minute, second) into a
        (year, month, day.fractional) date.

        :param tuple date: A six part date (y, m, d, hh, mm, ss).
        :returns: The three part (y, m, d.nnn).
        :rtype: tuple
        """
        self._check_valid_gregorian_month_day(date)
        t_len = len(date)
        year = date[0]
        month = date[1]
        day = date[2]
        hour = date[3] if t_len > 3 and date[3] is not None else 0
        minute = date[4] if t_len > 4 and date[4] is not None else 0
        second = date[5] if t_len > 5 and date[5] is not None else 0
        day += round(self._HR(hour) + self._MN(minute) + self._SEC(second),
                     self._ROUNDING_PLACES)
        return year, month, day

    def ymdhms_from_date(self, date: tuple, us: bool=False) -> tuple:
        """
        Convert (year, month, day.fractional) into a
        (year, month, day, hour, minute, second).

        :param tuple date: A three part date (y, m, d.nnn).
        :param bool us: If True return microseconds as seperate field from
                        seconds else return seconds with fractional seconds.
                        Default is False.
        :returns: A six part date (y, m, d, hh, mm, ss).
        :rtype: tuple
        """
        self._check_valid_gregorian_month_day(date)
        date_len = len(date)
        year = date[0]
        month = date[1]
        day = date[2]
        hour = date[3] if date_len > 3 else 0
        minute = date[4] if date_len > 4 else 0
        second = date[5] if date_len > 5 else 0
        microsec = date[6] if date_len > 6 else 0
        total_seconds = ((hour * 3600) + (minute * 60) + second +
                         (microsec / 1e6))
        day += total_seconds / 86400
        hhmmssus = self._hms_from_decimal_day(day, us=us)
        return (year, month, math.floor(day)) + hhmmssus

    def _check_valid_gregorian_month_day(self, g_date: tuple,
                                         historical: bool=False) -> None:
        """
        Check that the month and day values are valid.

        :param tuple g_date: The date to check.
        :param bool historical: If True use the Julian leap year before 1883.
        :returns: Nothing
        :rtype: None
        """
        t_len = len(g_date)
        year = g_date[0]
        month = abs(g_date[1])
        day = abs(g_date[2])
        LY = (self._JULIAN_LEAP_YEAR if historical and year < 1883
              else self._GREGORIAN_LEAP_YEAR)
        hour = g_date[3] if t_len > 3 and g_date[3] is not None else 0
        minute = g_date[4] if t_len > 4 and g_date[4] is not None else 0
        second = g_date[5] if t_len > 5 and g_date[5] is not None else 0
        assert 1 <= month <= 12, f"Invalid month '{month}', should be 1 - 12."
        days = self._MONTHS[month - 1]

        if month == 2:  # Subtract 0 or 1 from February if leap year.
            days += LY(year)

        assert 1 <= math.floor(day) <= days, (
            f"Invalid day '{day}' for month '{month}' and year '{year}' "
            f"should be 1 - {days}.")
        assert hour < 24, f"Invalid hour '{hour}' it must be < 24"
        assert minute < 60, f"Invalid minute '{minute}' should be < 60."

        if any((hour, minute, second)):
            assert not day % 1, ("If there is a part day then there can be no "
                                 "hours, minutes, or seconds.")

        if any((minute, second)):
            assert not hour % 1, (
                "If there is a part hour then there can be no minutes or "
                "seconds.")

        if second:
            assert not minute % 1, (
                "If there is a part minute then there can be no seconds.")



# -*- coding: utf-8 -*-
#
# badidatetime/base_calendar.py
#
__docformat__ = "restructuredtext en"

import math

from badidatetime.julian_period import JulianPeriod
from badidatetime._astronomical_terms import AstronomicalTerms


class BaseCalendar(AstronomicalTerms, JulianPeriod):
    """
    Basic functionality used with all calenders.

    U.T. = Mean solar time at Greenwich, England (0◦ meridian), reckoned from
           midnight; sometimes GMT, Greenwich Mean Time

    Transformations between Time Systems:
    | https://gssc.esa.int/navipedia/index.php/Transformations_between_Time_Systems
    """
    _HR = lambda self, x: x / 24
    _MN = lambda self, x: x / 24 / 60
    _SEC = lambda self, x: x / 24 / 60 / 60
    _MINS = lambda self, x: x / 60
    _SECS = lambda self, x: x / 3600
    # Convert microseconds to a partial second.
    _US = lambda self, x: x / 1000000
    _ANGLE = lambda self, d, m, s: d + (m + s / 60) / 60  # 0 - 360
    _AMOD = lambda self, x, y: y + x % -y
    _MOD3 = lambda self, x, a, b: x if a == b else (
        a + math.fmod((x - a), (b - a)))
    _QUOTIENT = lambda self, m, n: math.floor(m / n)

    # The inline functions below will assume that 0 is midnight, if
    # converting from a Julian Period day add 0.5 to the value before
    # calling the function.
    _PARTIAL_DAY_TO_HOURS = lambda self, x: round(
        (x % 1) * 24, self._ROUNDING_PLACES)
    _PARTIAL_HOUR_TO_MINUTE = lambda self, x: round(
        (x % 1) * 60, self._ROUNDING_PLACES)
    _PARTIAL_MINUTE_TO_SECOND = _PARTIAL_HOUR_TO_MINUTE
    _PARTIAL_SECOND_TO_MICROSECOND = lambda self, x: int(
        round(x % 1, self._ROUNDING_PLACES) * 1e6)

    _MEAN_TROPICAL_YEAR = 365.2421897
    # MEAN_SIDEREAL_YEAR = 365.256363004

    _MORNING = True
    _EVENING = False
    # Seasons set to degrees.
    _SPRING = 0
    _SUMMER = 90
    _AUTUMN = 180
    _WINTER = 270
    _SUN_OFFSET = 0.8333333333333334
    _STARS_PLANET_OFFSET = 0.5666666666666667
    _ROUNDING_PLACES = 6
    _POSIX_EPOCH = 2440585.5  # This is using the more exact algorithm.
    _JULIAN_CAL_EPOCH = 1721423.5

    def __init__(self, *args, **kwargs):
        self._time = None
        self._nutation = {'lon': (0, 0, False), 'obl': (0, 0, False)}
        self._sun_tss = {'trn': (), 'rsn': (), 'stn': ()}

    #
    # Meeus Astronomical Algorithms
    #

    def _delta_t(self, jd: float, *, seconds: bool=False) -> float:
        """
        Calculate the value of ΔT = TD − UT, for TD = ΔT + UT, and for
        UT = ΔT - TD. Only the year and month are considered, days, hours,
        minutes, and seconds are ignored.

        :param float jd: Julian day.
        :param bool seconds: If True leave as seconds in a minute else convert
                             to seconds of a day.
        :returns: The delta t.
        :rtype: float

        .. note::

           See: http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html
        """
        from .gregorian_calendar import GregorianCalendar
        gc = GregorianCalendar()
        g_date = gc.gregorian_date_from_jd(jd)
        year = g_date[0] + (g_date[1] - 0.5) / 12
        func = lambda year: -20 + 32 * ((year - 1820) / 100)**2

        if year < -500:
            dt = func(year)
        elif year < 500:
            u = year / 100
            dt = self._poly(u, (10583.6, -1014.41, 33.78311, -5.952053,
                                -0.1798452, 0.022174192, 0.0090316521))
        elif year < 1600:
            u = (year - 1000) / 100
            dt = self._poly(u, (1574.2, -556.01, 71.23472, 0.319781,
                                -0.8503463, -0.005050998, 0.0083572073))
        elif year < 1700:
            u = year - 1600
            dt = self._poly(u, (120, -0.9808, -0.01532, 1 / 7129))
        elif year < 1800:
            u = year - 1700
            dt = self._poly(u, (8.83, 0.1603, -0.0059285, 0.00013336,
                                -1 / 1174000))
        elif year < 1860:
            u = year - 1800
            dt = self._poly(u, (13.72, -0.332447, 0.0068612, 0.0041116,
                                -0.00037436, 0.0000121272, -0.0000001699,
                                0.000000000875))
        elif year < 1900:
            u = year - 1860
            dt = self._poly(u, (7.62, 0.5737, -0.251754, 0.01680668,
                                -0.0004473624, 1 / 233174))
        elif year < 1920:
            u = year - 1900
            dt = self._poly(u, (-2.79, 1.494119, -0.0598939, 0.0061966,
                                -0.000197))
        elif year < 1941:
            u = year - 1920
            dt = self._poly(u, (21.20, 0.84493, -0.0761, 0.0020936))
        elif year < 1961:
            u = year - 1950
            dt = self._poly(u, (29.07, 0.407, -1 / 233, 1 / 2547))
        elif year < 1986:
            u = year - 1975
            dt = self._poly(u, (45.45, 1.067, -1 / 260, -1 / 718))
        elif year < 2005:
            u = year - 2000
            dt = self._poly(u, (63.86, 0.3345, -0.060374, 0.0017275,
                                0.000651814, 0.00002373599))
        elif year < 2050:
            u = year - 2000
            dt = self._poly(u, (62.92, 0.32217, 0.005589))
        elif year < 2150:
            dt = -20 + 32 * ((year - 1820) / 100)**2 - 0.5628 * (2150 - year)
        else:  # 2150 >= year
            dt = func(year)

        # Convert to seconds of a day where 66.9 dt == 2010
        # Seconds of a day are 0.0007743055555555556
        return dt if seconds else dt / 86400

    def _mean_sidereal_time_greenwich(self, tc: float) -> float:
        """
        Mean sidereal time at Greenwich. (GMST) If the hour angle is
        measured with respect to the mean equinox, mean sidereal time is
        being measured.

        :param float tc: Julian century.
        :returns: The mean sidereal time at Greenwich .
        :rtype: float

        .. note::

           Meeus--AA ch.12 p.88 Eq.12.3
        """
        return self._coterminal_angle(280.46061837 + 360.98564736629 *
                                      (tc * 36525) + 0.000387933 *
                                      tc**2 - tc**3 / 38710000)

    def _apparent_sidereal_time_greenwich(self, tc: float) -> float:
        """
        The apparent sidereal time, or the Greenwich. (GAST) If the hour
        angle is measured with respect to the true equinox, apparent
        sidereal time is being measured.

        :param float tc: Julian century.
        :returns: The apparent sidereal time at Greenwich.
        :rtype: float

        .. note::

           Meeus--AA ch.12 p.88
        """
        t0 = self._mean_sidereal_time_greenwich(tc)
        eps = self._true_obliquity_of_ecliptic(tc)
        d_psi = self._nutation_longitude(tc)
        return self._coterminal_angle(t0 + d_psi * self._cos_deg(eps))

    def _altitude(self, delta: float, lat: float, h: float) -> float:
        """
        Altitude in degrees, positive above the horizon, negative below.

        :param float delta: Declination in sidereal time.
        :param float lat: Geographic latitude.
        :param float h: Local hour angle.
        :returns: Altitude in degrees.
        :rtype: float

        .. note::

           Meeus--AA p.93 Eq.13.6
        """
        return math.degrees(math.asin(
            self._sin_deg(lat) * self._sin_deg(delta) + self._cos_deg(lat) *
            self._cos_deg(delta) * self._cos_deg(h)))

    def _approx_local_hour_angle(self, tc: float, lat: float,
                                 offset: float=_SUN_OFFSET) -> float:
        """
        Approximate local hour angle, measured westwards from the south
        in degrees.

        Hour angle, in astronomy, the angle between an observer’s meridian
        (a great circle passing over his head and through the celestial
        poles) and the hour circle (any other great circle passing through
        the poles) on which some celestial body lies. This angle, when
        expressed in hours and minutes, is the time elapsed since the
        celestial body’s last transit of the observer’s meridian.

        :param float tc: Julian century.
        :param float lat: Latitude in decimal
        :param float offset: A constant “standard” altitude, i.e., the
                             geometric altitude of the center of the body at
                             the time of apparent rising or setting, namely,
                             h0 = -0°34’ = -0°5667 for stars and planets;
                             h0 = -0°50' = -0°8333 for the Sun.
                             Default is _SUN_OFFSET, _STARS_PLANET_OFFSET can
                             also be used.
        :returns: The approximat local hour angle in degrees.
        :rtype: float

        .. note::

           1. Meeus--AA p.101,102
           2. If the result of the equation is negative then add 360°
              (6.283185307179586 radians). If result is greater than
              360° then subtract 360° (6.283185307179586 radians).
              https://www.quora.com/How-do-I-calculate-the-hour-angle
        """
        delta = self._sun_apparent_declination(tc)
        cos_h0 = ((self._sin_deg(-offset) - self._sin_deg(lat) *
                   self._sin_deg(delta)) / (self._cos_deg(lat) *
                                            self._cos_deg(delta)))

        if cos_h0 < -1 or cos_h0 > 1:
            cos_h0 -= math.floor(cos_h0)

        return math.degrees(math.acos(cos_h0))

    def _sun_transit(self, jd: float, lon: float, zone: float=0.0,
                     exact_tz: bool=False) -> float:
        """
        The transit is when the body crosses the local maridian at upper
        culmination.

        :param float jd: Julian day in UT.
        :param float lon: Geographic longitude positive east negative west.
        :param float zone: This is the political timezone.
        :param bool exact_tz: Derive the timezone from the longitude. The
                              'zone' parameter is not used if this is True,
                              Default is False.
        :returns: The center point between sunrise and sunset.
        :rtype: float

        .. note::

           Meeus-AA ch.15 p. 102, 103 Eq.15.1, 15.2
        """
        assert ((-180 <= zone <= 180 and not exact_tz) or
                (zone == 0 and exact_tz)), (
            f"If exact_tz is True then zone must be 0, found zone: {zone} "
            f"and exact_tz: {exact_tz}.")
        zone = zone if not exact_tz else lon / 15
        func0 = lambda m: m + 1 if m <= 0 else m - 1 if m >= 1 else m
        tc = self._julian_centuries(jd)
        dt = self._delta_t(jd)
        tc_td = dt / 36525 + tc  # Compensate for the Julian Century
        alpha = self._sun_apparent_right_ascension(tc_td)
        ast = self._apparent_sidereal_time_greenwich(tc)
        m = func0((alpha - lon - ast) / 360)
        md = self._transit_correction(tc, ast, dt, lon, m)
        m += md + self._tz_decimal_from_dhms(0, zone, 0, 0)
        return m

    def _transit_correction(self, tc: float, ast: float, dt: float, lon: float,
                            m: float) -> float:
        """
        Find the correction to the transit.

        :param float tc: The Julian Period century.
        :param float ast: The apparent sidereal time at greenwich.
        :param float dt: The delta T of the JD.
        :param float lon: The longitude.
        :param float m: Times on day, expressed as fractions.
        :returns: The correction to the transit.
        :rtype: float
        """
        srt = ast + 360.98564736629 * m
        n = m + dt / 86400
        ra0 = self._sun_apparent_right_ascension(tc - (1 / 36525))
        ra1 = self._sun_apparent_right_ascension(tc)
        ra2 = self._sun_apparent_right_ascension(tc + (1 / 36525))
        alpha = self._interpolation_from_three(ra0, ra1, ra2, n, True)
        h = self._local_hour_angle(srt, lon, alpha)
        return -h / 360

    def _sun_rising(self, jd: float, lat: float, lon: float, zone: float=0, *,
                    exact_tz: bool=False, offset: float=_SUN_OFFSET) -> float:
        """
        Find the jd for sunrise of the given jd.

        :param float jd: Julian day in UT.
        :param float lat: Geographic latitude positive north negative south.
        :param float lon: Geographic longitude positive east negative west.
        :param float zone: The time zone, defaults to the zero zone in
                           Greenwich UK.
        :param bool exact_tz: The political time zones or the exact time zone
                              derived from the longitude
                              (15 degrees = 360 / 24).
                              Default is False or the political time zone.
        :param bool offset: A constant “standard” altitude, i.e., the geometric
                            altitude of the center of the body at the time of
                            apparent rising or setting, namely,
                            h0 = -0°34’ = -0°5667 for stars and planets;
                            h0 = -0°50' = -0°8333 for the Sun.
                            Default is _SUN_OFFSET, _STARS_PLANET_OFFSET can
                            also be used.
        :returns: The jd moments of the sunrise.
        :rtype: float

        .. note::

           Meeus-AA ch.15 p. 102, 103 Eq.15.1, 15.2
        """
        jd += self._rising_setting(jd, lat, lon, zone, exact_tz=exact_tz,
                                   offset=offset, sr_ss='RISE')
        return round(jd, self._ROUNDING_PLACES)

    def _sun_setting(self, jd: float, lat: float, lon: float, zone: float=0, *,
                     exact_tz: bool=False, offset: float=_SUN_OFFSET) -> float:
        """
        Find the jd for sunset of the given jd.

        :param float jd: Julian day in UT.
        :param float lat: Geographic latitude positive north negative south.
        :param float lon: Geographic longitude positive east negative west.
        :param float zone: The time zone, defaults to the zero zone in
                           Greenwich UK.
        :param bool exact_tz: The political time zones (default) or the exact
                              time zone. The exact time zone is derived from
                              the longitude (15 degrees = 360 / 24).
        :param bool offset: A constant “standard” altitude, i.e., the geometric
                            altitude of the center of the body at the time of
                            apparent rising or setting, namely,
                            h0 = -0°34’ = -0°5667 for stars and planets;
                            h0 = -0°50' = -0°8333 for the Sun.
                            Default is _SUN_OFFSET, _STARS_PLANET_OFFSET can
                            also be used.
        :returns: The jd moments of the sunset.
        :rtype: float

        .. note::

           Meeus-AA ch.15 p. 102, 103 Eq.15.1, 15.2
        """
        jd += self._rising_setting(jd, lat, lon, zone, exact_tz=exact_tz,
                                   offset=offset, sr_ss='SET')
        return round(jd, self._ROUNDING_PLACES)

    def _rising_setting(self, jd: float, lat: float, lon: float, zone: float=0,
                        *, exact_tz: bool=False, offset: float=_SUN_OFFSET,
                        sr_ss: str='RISE') -> float:
        """
        Find the jd difference for sunrise or sunset of the given jd.

        :param float jd: Julian day in UT.
        :param float lat: Geographic latitude positive north negative south.
        :param float lon: Geographic longitude positive east negative west.
        :param float zone: The time zone, defaults to the zero zone in
                           Greenwich UK.
        :param bool exact_tz: The political time zones or the exact time zone
                              derived from the longitude
                              (15 degrees = 360 / 24).
                              Default is False or the political time zone.
        :param float offset: A constant “standard” altitude, i.e., the
                             geometric altitude of the center of the body at
                             the time of apparent rising or setting, namely,
                             h0 = -0°34’ = -0°5667 for stars and planets;
                             h0 = -0°50' = -0°8333 for the Sun.
                             Default is _SUN_OFFSET, _STARS_PLANET_OFFSET can
                             also be used.
        :param str sr_ss: If 'RISE' return the sunrise else return sunset.
        :returns: The offset that would be added to the currect date.
        :rtype: float

        .. note::

           Meeus-AA ch.15 p. 102, 103 Eq.15.1, 15.2
        """
        assert ((-180 <= zone <= 180 and not exact_tz) or
                (zone == 0 and exact_tz)), (
            f"If exact_tz is True then zone must be 0, found zone: {zone} "
            f"and exact_tz: {exact_tz}.")
        zone = zone if not exact_tz else lon / 15
        flags = ('RISE', 'SET')
        sr_ss = sr_ss.upper()
        assert sr_ss in flags, (
            f"Invalid value, should be one of '{flags}' found '{sr_ss}'.")
        func0 = lambda m: m + 1 if m <= 0 else m - 1 if m >= 1 else m
        tc = self._julian_centuries(jd)
        dt = self._delta_t(jd)
        tc_td = dt / 36525 + tc  # Compensate for the Julian Century
        alpha = self._sun_apparent_right_ascension(tc_td)
        ast = self._apparent_sidereal_time_greenwich(tc)
        h0 = self._approx_local_hour_angle(tc, lat, offset=offset)
        m0 = func0((alpha - lon - ast) / 360)
        m = m0 - h0 / 360 if sr_ss == 'RISE' else m0 + h0 / 360
        dm = 1

        for i in range(5):
            dm = self._rise_set_correction(tc, ast, dt, lat, lon, m, offset)
            m += dm
            if abs(dm) < 0.0001: break

        m += self._tz_decimal_from_dhms(0, zone, 0, 0)
        return m % 1

    def _rise_set_correction(self, tc: float, ast: float, dt: float,
                             lat: float, lon: float, m: float, offset: float
                             ) -> float:
        """
        Find the correction to the sunrise and sunset.

        :param float tc: The Julian Period century.
        :param float ast: The apparent sidereal time at greenwich.
        :param float dt: The delta T of the JD.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float m: Times on day, expressed as fractions.
        :param float offset: Either `_SUN_OFFSET` or `_STARS_PLANET_OFFSET`.
        :returns: The correction to the sunrise or sunset.
        :rtype: float
        """
        srt = ast + 360.98564736629 * m
        n = m + dt / 86400
        ra0 = self._sun_apparent_right_ascension(tc - (1 / 36525))
        ra1 = self._sun_apparent_right_ascension(tc)
        ra2 = self._sun_apparent_right_ascension(tc + (1 / 36525))
        de0 = self._sun_apparent_declination(tc - (1 / 36525))
        de1 = self._sun_apparent_declination(tc)
        de2 = self._sun_apparent_declination(tc + (1 / 36525))
        alpha = self._interpolation_from_three(ra0, ra1, ra2, n, True)
        delta = self._interpolation_from_three(de0, de1, de2, n)
        h = self._local_hour_angle(srt, lon, alpha)
        alt = self._altitude(delta, lat, h)
        return (alt + offset) / (360 * self._cos_deg(delta) *
                                 self._cos_deg(lat) * self._sin_deg(h))

    def _local_hour_angle(self, srt: float, lon: float, alpha: float) -> float:
        """
        The LHA is the angle between the meridian of the observer and the
        meridian of the geographical position of the celestial body.

        :param float srt: Sidereal time.
        :param float lon: The Geographic longitude of the observer in degrees.
        :param float alpha: The apparent right ascensions.
        :returns: The local hour angle.
        :rtype: float

        .. note::

           | Meeus-AA p.103
           | https://astronavigationdemystified.com/local-hour-angle-and-greenwich-hour-angle/

        """
        h = self._coterminal_angle(srt + lon - alpha)
        return h - 360 if h > 180 else h

    def _nutation_longitude(self, tc: float, *, degrees: bool=False) -> float:
        """
        Nutation longitude of the Earth's axis around it's 'mean' position.

        :param float tc: The Julian Period century.
        :param bool degrees: If `False` (default) return radians else if
                             `True` return degrees.
        :returns: Either the radians or degrees depending on the `degrees`
                  argument.
        :rtype: float
        """
        tc_day, lon_sum, deg = self._nutation['lon']

        if tc != tc_day or degrees != deg:
            lon_sum, obl_sum = self._nutation_obliquity_longitude(
                tc, degrees=degrees)
            self._nutation['lon'] = (tc, lon_sum, degrees)
            self._nutation['obl'] = (tc, obl_sum, degrees)

        return self._nutation['lon'][1]

    def _nutation_obliquity(self, tc: float, *, degrees: bool=False) -> float:
        """
        Nutation obliquity of the Earth's equator around it's 'mean' position.

        :param float tc: The Julian Period century.
        :param bool degrees: If `False` (default) return radians else if
                             `True` return degrees.
        :returns: The nutation of obliquity.
        :rtype: float
        """
        tc_day, obj_sum, deg = self._nutation['obl']

        if tc != tc_day or degrees != deg:
            lon_sum, obl_sum = self._nutation_obliquity_longitude(
                tc, degrees=degrees)
            self._nutation['lon'] = (tc, lon_sum, degrees)
            self._nutation['obl'] = (tc, obl_sum, degrees)

        return self._nutation['obl'][1]

    def _nutation_obliquity_longitude(self, tc: float, degrees: bool=False
                                      ) -> float:
        """
        Nutation of the Earth's axis around it's 'mean' position.

        :param float tc: Time in Julian centuries.
        :param float degrees: If True units of degrees is returned, if False
                              units of radians is returned.
        :returns: Moon latitude.
        :rtype: float

        .. note::

           See: https://articles.adsabs.harvard.edu/full/seri/CeMec/0027/0000079.000.html
        """
        lm = self._moon_mean_anomaly(tc)
        ls = self._sun_earth_mean_anomaly(tc)
        ff = self._moon_latitude(tc)
        dd = self._mean_moon_elongation(tc)
        om = self._moon_ascending_node_longitude(tc)

        # W = LM*lm + LS*ls + F*ff + D*dd + OM*om
        # T = tc
        # Where LM, LS, F, D, and OM are from the nutation periodic terms.
        # The nutation in longitude is a sum of terms of the form
        # (psi_sin + sin * T) * sin(W).
        # Where psi_sin and t_sin are from the nutation periodic terms
        # and T is the Julian time from J2000.
        # The obliquity in latitude is a sum of terms of the form
        # (eps_cos + cos * T) * cos(W).
        # Where eps_cos and t_cos are from the nutation periodic terms
        # and T is the Julian time from J2000.
        lon_sum = 0
        obl_sum = 0

        for LM, LS, F, D, OM, day, psi_sin, sin, eps_cos, cos in self._NUT:
            w = LM*lm + LS*ls + F*ff + D*dd + OM*om
            lon_sum += (psi_sin + sin * tc) * self._sin_deg(w)
            obl_sum += (eps_cos + cos * tc) * self._cos_deg(w)

        lon_sum /= 36000000
        obl_sum /= 36000000

        if degrees:
            lon_sum = self._coterminal_angle(math.degrees(lon_sum))
            obl_sum = self._coterminal_angle(math.degrees(obl_sum))

        return lon_sum, obl_sum

    def _moon_mean_anomaly(self, tc: float) -> float:
        """
        The position of the moon in its orbit around the Earth.

        :param float tc: Time in Julian centuries.
        :returns: Moon mean anomaly.
        :rtype: float

        .. note::

           | Meeus--AA ch.22 p.144
           | Referenced by lm (M').
        """
        return self._coterminal_angle(self._poly(
            tc, (134.96298, 477198.867398, 0.0086972, 1 / 56250)))

    def _sun_earth_mean_anomaly(self, tc: float) -> float:
        """
        The position of the Earth in its orbit around the Sun.

        :param float tc: Time in Julian centuries.
        :returns: Sun and earth mean anomaly.
        :rtype: float

        .. note::

           | Meeus--AA ch.22 p.144
           | Referenced by ls (M).
        """
        return self._coterminal_angle(self._poly(
            tc, (357.52772, 35999.05034, -0.0001603, -1 / 300000)))

    def _moon_latitude(self, tc: float) -> float:
        """
        The angle between the Moon’s ecliptic longitude and its mean
        longitude.

        :param float tc: Time in Julian centuries.
        :returns: Moon latitude.
        :rtype: float

        .. note::

           | Meeus--AA ch.22 p.144
           | Referenced by ff (F).
        """
        return self._coterminal_angle(self._poly(
            tc, (93.27191, 483202.017538, -0.0036825, 1 / 327270)))

    def _mean_moon_elongation(self, tc: float) -> float:
        """

        :param float tc: Time in Julian centuries.
        :returns: Mean moon elongation.
        :rtype: float

        .. note::

           | Meeus--AA ch22 p144
           | Referenced by dd (D).
        """
        return self._coterminal_angle(self._poly(
            tc, (297.85036, 445267.11148, -0.0019142, 1 / 189474)))

    def _moon_ascending_node_longitude(self, tc: float) -> float:
        """
        Longitude of the ascending node of the Moon’s mean orbit on the
        ecliptic, measured from the mean equinox of the date:

        :param float tc: Time in Julian centuries.
        :returns: Moon ascending node longitude.
        :rtype: float

        .. note::

           | Meeus--AA ch22 p144
           | Referenced by om (omega).
        """
        return self._coterminal_angle(self._poly(
            tc, (125.04452, -1934.136261, 0.0020708, 1 / 450000)))

    def _true_obliquity_of_ecliptic(self, tc: float) -> float:
        """
        The obliquity of the ecliptic, or inclination of the Earth’s axis
        of rotation, is the angle between the equator and the ecliptic.

        :param float jde: Julian century.
        :returns: The obliquity of the ecliptic in degrees as a decimal.
        :rtype: float

        .. note::

           | Meeus--AA ch.22 p.147 Eq.22.3
           | Convert lots of things:
           | https://www.xconvert.com/unit-converter/arcseconds-to-degrees
        """
        u = tc / 100
        mean_ob = self._poly(u, (23.43929111111111, -1.3002583333333335,
                                 -0.00043055555555555555, 0.5553472222222222,
                                 -0.014272222222222223, -0.06935277777777778,
                                 -0.010847222222222222, 0.001977777777777778,
                                 0.007741666666666667, 0.0016083333333333334,
                                 0.0006805555555555556))
        return mean_ob + self._nutation_obliquity(tc)

    def _sun_mean_longitude(self, tc: float) -> float:
        """
        The geometric mean longitude of the Sun, referred to the mean
        equinox of the date.

        :param float tc: Time in Julian centuries.
        :returns: Mean longitude of the sun in degrees.
        :rtype: float

        .. note::

           | Meeus--AA ch.25 p.163 Eq.25.2
           | References by L0
        """
        return self._coterminal_angle(self._poly(
            tc, (280.46646, 36000.76983, 0.0003032)))

    def _eccentricity_earth_orbit(self, tc: float) -> float:
        """
        The eccentricity of the earth's orbit.

        :param float tc: Time in Julian centuries.
        :returns: The eccentricity of the earth's orbit.
        :rtype: float

        .. note::

           Meeus--AA ch.25 p.163 Eq.25.4
        """
        return self._poly(tc, (0.016708634, -0.000042037, -0.0000001267))

    def _sun_equation_of_center(self, tc: float) -> float:
        """
        Describes the difference between the true anomaly (the actual
        angular position of the Earth in its orbit around the Sun) and the
        mean anomaly (the position the Earth would have if its orbit were
        perfectly circular and uniform). This difference is caused by the
        elliptical shape of Earth’s orbit and the resulting non-uniform
        motion of the Earth around the Sun.

        :param float tc: Time in Julian centuries.
        :returns: The Sun's equation of it's center.
        :rtype: float

        .. note::

           Meeus--AA ch.25 p.164
        """
        m = self._sun_earth_mean_anomaly(tc)
        return ((1.914602 - 0.004817 * tc - 0.000014 * tc**2) *
                self._sin_deg(m) + (0.019993 - 0.000101 * tc) *
                self._sin_deg(2 * m) + 0.000290 * self._sin_deg(3 * m))

    def _sun_true_longitude(self, tc: float) -> float:
        """
        The true geometric longitude referred to the mean equinox of the
        date. This longitude is the quantity required for instance in the
        calculation of geocentric planetary positions.

        :param float tc: Time in Julian centuries.
        :returns: The true geometric longitude.
        :rtype: float

        .. note::

           Meeus--AA p.164
        """
        l0 = self._sun_mean_longitude(tc)
        cen = self._sun_equation_of_center(tc)
        return l0 + cen

    def _sun_apparent_longitude(self, tc: float) -> float:
        """
        The Sun’s apparent longitude is the angle measured from the vernal
        equinox to the Sun’s position on the ecliptic plane as seen from
        Earth.

        :param float tc: Time in Julian centuries.
        :returns: The Sun’s apparent longitude.
        :rtype: float

        .. note::

           Meeus--AA p.164

           This has a less accurate result. apparent_solar_longitude()
           should be more acurate.
        """
        sol = self._sun_true_longitude(tc)
        om = self._moon_ascending_node_longitude(tc)
        return sol - 0.00569 - 0.00478 * self._sin_deg(om)

    def _sun_apparent_right_ascension(self, tc: float) -> float:
        """
        Right ascension is measured (from 0 to 24 hours, sometimes from 0°
        to 360°) from the vernal equinox, positive to the east, along the
        celestial equator.

        :param float tc: Julian century.
        :returns: The apparent declination of the sun in radians.
        :rtype: float

        .. note::

           Meeus--AA ch.25 p.165 Eq.25.6
        """
        om = self._moon_ascending_node_longitude(tc)
        eps = (self._true_obliquity_of_ecliptic(tc) + 0.00256 *
               self._cos_deg(om))
        lam = self._sun_apparent_longitude(tc)
        alpha = math.degrees(math.atan2(self._cos_deg(eps) *
                                        self._sin_deg(lam),
                                        self._cos_deg(lam)))
        return self._coterminal_angle(alpha)

    def _sun_apparent_declination(self, tc: float) -> float:
        """
        Declination is measured (from 0° to +90°) from the equator, positive
        to the north, negative to the south.

        :param float tc: Julian century.
        :param bool app: If True the apparent declination is returned, if
        :returns: The apparent declination of the sun in radians.
        :rtype: float

        .. note::

           Meeus--AA ch.25 p165 Eq25.7
        """
        om = self._moon_ascending_node_longitude(tc)
        eps = (self._true_obliquity_of_ecliptic(tc) + 0.00256 *
               self._cos_deg(om))
        lam = self._sun_apparent_longitude(tc)
        return math.degrees(math.asin(self._sin_deg(eps) * self._sin_deg(lam)))

    def _heliocentric_ecliptical_longitude(self, tm: float,
                                           degrees: bool=False) -> float:
        """
        Find the heliocentric ecliptical longitude.

        :param float tm: The moment in time referenced to J2000 millennia.
        :param bool degrees: The results if False are radians, else True
                             are degrees. Default is False.
        :returns: Longitude in degrees or radians.
        :rtype: float

        .. note::

           | Meeus--AA ch.25 p.166
           | Referenced by L
        """
        func = lambda a, b, c: a * math.cos(b + c * tm)
        l0 = self._sigma((self._L0_A, self._L0_B, self._L0_C), func)
        l1 = self._sigma((self._L1_A, self._L1_B, self._L1_C), func)
        l2 = self._sigma((self._L2_A, self._L2_B, self._L2_C), func)
        l3 = self._sigma((self._L3_A, self._L3_B, self._L3_C), func)
        l4 = self._sigma((self._L4_A, self._L4_B, self._L4_C), func)
        l5 = self._sigma((self._L5_A, self._L5_B, self._L5_C), func)
        l = self._poly(tm, (l0, l1, l2, l3, l4, l5)) / 10**8
        return round(self._coterminal_angle(math.degrees(l)) if degrees else l,
                     self._ROUNDING_PLACES)

    def _heliocentric_ecliptical_latitude(self, tm: float,
                                          degrees: bool=False) -> float:
        """
        Find the heliocentric ecliptical latitude.

        :param float tm: The moment in time referenced to J2000 millennia.
        :param bool degrees: The results if False are radians, else True
                             are degrees. Default is False.
        :returns: Latitude in degrees or radians.
        :rtype: float

        .. note::

           | Meeus--AA ch.25 p.166
           | Referenced by B
        """
        func = lambda a, b, c: a * math.cos(b + c * tm)
        b0 = self._sigma((self._B0_A, self._B0_B, self._B0_C), func)
        b1 = self._sigma((self._B1_A, self._B1_B, self._B1_C), func)
        b = self._poly(tm, (b0, b1)) / 10**8
        return round(self._coterminal_angle(math.degrees(b)) if degrees else b,
                     self._ROUNDING_PLACES)

    def _radius_vector(self, tm: float, degrees: bool=False) -> float:
        """
        Find the distance of earth to the sun.

        :param float tm: The moment in time referenced to J2000 millennia.
        :param bool degrees: The results if False are radians, else True
                             are degrees. Default is False.
        :returns: Radius vector in degrees or radians.
        :rtype: float

        .. note::

           | Meeus--AA ch.25 p.166
           | Referenced by R
        """
        func = lambda a, b, c: a * math.cos(b + c * tm)
        r0 = self._sigma((self._R0_A, self._R0_B, self._R0_C), func)
        r1 = self._sigma((self._R1_A, self._R1_B, self._R1_C), func)
        r2 = self._sigma((self._R2_A, self._R2_B, self._R2_C), func)
        r3 = self._sigma((self._R3_A, self._R3_B, self._R3_C), func)
        r4 = self._sigma((self._R4_A, self._R4_B, self._R4_C), func)
        r = self._poly(tm, (r0, r1, r2, r3, r4)) / 10**8
        return round(self._coterminal_angle(math.degrees(r)) if degrees else r,
                     self._ROUNDING_PLACES)

    def _apparent_solar_longitude(self, jde: float,
                                  degrees: bool=True) -> float:
        """
        Find the apparent solar longitude.

        :param float jde: The Julian Period day.
        :param bool degrees: The results if False are radians, else True
                             are degrees. Default is False.
        :returns: The apparent solar longitude.
        :rtype: float

        .. note::

           Meeus--AA ch.25 p.166
        """
        tm = self._julian_millennia(jde)
        l = self._heliocentric_ecliptical_longitude(tm, degrees=False)
        l += math.pi
        # Convert to FK5 notation
        l -= math.radians(2.5091666666666666e-05)  # -0".09033
        # Convert to centuries
        l += self._nutation_longitude(tm * 10, degrees=False)
        # eq 25.11
        l += self._aberration(tm)

        if degrees:
            l = self._coterminal_angle(math.degrees(l))

        return round(l, self._ROUNDING_PLACES)

    def _apparent_solar_latitude(self, jde: float,
                                 degrees: bool=True) -> float:
        """
        Find the apparent solar latitude.

        :param float jde: The Julian Period day.
        :param bool degrees: The results if False are radians, else True
                             are degrees. Default is False.
        :returns: The apparent solar latitude.
        :rtype: float

        .. note::

           Meeus--AA ch.25 p.166
        """
        tm = self._julian_millennia(jde)
        tc = tm * 10  # Convert millenna to centuries
        b = self._heliocentric_ecliptical_latitude(tm)
        b *= -1  # Invert the result
        # Convert to FK5 notation
        l = self._apparent_solar_longitude(jde, degrees=False)
        b1 = self._poly(tc, (l, -1.397, -0.00031))
        bd = (math.radians(0.000010877777777777778) *
              (math.cos(b1) - math.sin(b1)))
        b += bd

        if degrees:
            b = self._coterminal_angle(math.degrees(b))

        return round(b, self._ROUNDING_PLACES)

    def _aberration(self, tm: float, fixed: bool=True) -> float:
        """
        Find the aberration of a date with respect to a fixed reference
        frame or to the mean equinox.

        :param float tc: The moment in time referenced to J2000 millennia.
        :param bool fixed: If `True` (default) the results is to a fixed
                           reference frame, if `False` the result is
                           referenced to the mean equinox.
        :returns: The aberration of the date in degrees.
        :rtype: float

        .. note::

           AA p.167 Eq.25.11, p.168
        """
        if fixed:
            aberration = 3548.193
        else:
            aberration = 3548.33

        for a, b, c in self._ABER_A:
            aberration += math.radians(a) * self._sin_deg(b + c * tm)

        for a, b, c in self._ABER_B:
            aberration += math.radians(a) * tm * self._sin_deg(b + c * tm)

        for a, b, c in self._ABER_C:
            aberration += math.radians(a) * tm**2 * self._sin_deg(b + c * tm)

        r = self._radius_vector(tm, degrees=False)
        return round(self._decimal_from_dms(
            0, 0, -0.005775518 * r * aberration), self._ROUNDING_PLACES)

    def _approx_julian_day_for_equinoxes_or_solstices(self, g_year: int,
                                                      lam: int=_SPRING
                                                      ) -> float:
        """
        Find the approximate Julian day for the equinoxes or solstices.

        :param int g_year: The Gregorian year.
        :param int lam: The lamda, either `_SPRING` (default), `_SUMMER`,
                        `_AUTUMN`, or `_WINTER`.
        :returns: The approximate Julian day for the equinoxes or solstices.
        :rtype: float

        .. note::

           See: Meeus AA ch.27 p.177
        """
        if g_year <= 1000:
            y = g_year / 1000

            if lam == self._SPRING:
                jde = self._poly(y, (1721139.29189, 365242.1374, 0.06134,
                                     0.00111, -0.00071))
            elif lam == self._SUMMER:
                jde = self._poly(y, (1721233.25401, 365241.72562, -0.05323,
                                     0.00907, 0.00025))
            elif lam == self._AUTUMN:
                jde = self._poly(y, (1721325.70455, 365242.49558, -0.11677,
                                     -0.00297, 0.00074))
            else:  # lam == self._WINTER:
                jde = self._poly(y, (1721414.39987, 365242.88257, -0.00769,
                                     -0.00933, -0.00006))
        else:
            y = (g_year - 2000) / 1000

            if lam == self._SPRING:
                jde = self._poly(y, (2451623.80984, 365242.37404, 0.05169,
                                     -0.00411, -0.00057))
            elif lam == self._SUMMER:
                jde = self._poly(y, (2451716.56767, 365241.62603, 0.00325,
                                     0.00888, -0.0003))
            elif lam == self._AUTUMN:
                jde = self._poly(y, (2451810.21715, 365242.01767, -0.11575,
                                     0.00337, 0.00078))
            else:  # lam == self._WINTER:
                jde = self._poly(y, (2451900.05952, 365242.74049, -0.06223,
                                     -0.00823, 0.00032))

        # JD♈(Y) = JD0+78.814+365.24236 ΔY+5.004*10^−8 ΔY^2−2.87*10^−12
        # ΔY^3−4.5*10^−16 ΔY^4

        return jde

    def _find_moment_of_equinoxes_or_solstices(self, jd: float,
                                               lam: int=_SPRING,
                                               zone: float=0) -> float:
        """
        With the jd and time of year find an equinoxe or solstice at
        Greenwich.

        :param float jd: Meeus algorithm Julian day.
        :param int lam: The lamda, either `_SPRING` (default), `_SUMMER`,
                        `_AUTUMN`, or `_WINTER`.
        :param float zone: The time zone.
        :returns: The Julian day of the equinox or solstice.
        :rtype: float

        .. note::

           Meeus--AA ch.27 p.177
        """
        from .gregorian_calendar import GregorianCalendar
        gc = GregorianCalendar()
        year = gc.gregorian_year_from_jd(jd)
        jde = self._approx_julian_day_for_equinoxes_or_solstices(year, lam)
        tc = self._julian_centuries(jde)
        w = 35999.373 * tc - 2.47
        dl = 1 + 0.0334 * self._cos_deg(w + 0.0007) * self._cos_deg(2*w)
        s = self._sigma((self._EQ_SO_A, self._EQ_SO_B, self._EQ_SO_C),
                        lambda a, b, c: a * self._cos_deg(b + c * tc))
        jde += (0.00001 * s) / dl + self._HR(zone)
        return round(jde, self._ROUNDING_PLACES)

    def _decimal_from_dms(self, degrees: int, minutes: int, seconds: float,
                          direction: str='N') -> float:
        '''
        Coordinantes in degrees, minutes, and seconds.

        | The Shrine of Baha’u’llah: 32°56’36.86″N, 35° 5’30.38″E
        | The Shrine of The Bab: 32°48’52.49″N, 34°59’13.91″E
        | The Guardian’s Resting Place (not 3D): 51°37’21.85″N, 0°08’35.57″W


        .. rubric:: Process

        Convert degrees, minutes, and seconds to a decimal.

        Degrees, minutes, and seconds to a decimal coordinant:

        1. Add the degrees to the minutes divided by 60
        2. Add the seconds divided by (60 x 60), which is 3600

        .. rubric:: Example

        To convert 35° 20′ 35", the answer is
        35 + (20/60) + (35/3600) = 35.34306 degrees.

        :param int degrees: The degree part of the coordinats.
        :param int minutes: The minute part of the coordinate.
        :param float seconds: The second part of the coordinate.
        :param str direction: The direction part of the coordinate which can
                              be any of the following N, S, E, W in upper or
                              lower case.
        :returns: latitude and longitude
        :rtype: tuple
        '''
        dirs = ('N', 'S', 'E', 'W')
        assert direction.upper() in dirs, (
            f"The 'direction' argument must be one of {dirs}")
        # Remove the minus sign iof it exists.
        degrees = -degrees if degrees < 0 else degrees
        decimal = degrees + (minutes / 60) + (seconds / 3600)
        # Adjust the sign based on the direction.
        return -decimal if direction.upper() in ('S', 'W') else decimal

    def _dms_from_decimal(self, coord: float, direction: str) -> tuple:
        """
        Convert a decimal degree into degrees, minutes, and seconds.

        :param float coord: The decimal coordinant.
        :param str direction: The direction part of the coordinate which can
                              be any of the following N, S, E, W in upper or
                              lower case.
        :returns: The degree, minute, second, and direction for of the
                  coordinate.
        :rtype: tuple
        """
        dirs = ('LATITUDE', 'LONGITUDE')
        direction = direction.upper()
        size = len(direction)
        assert size >= 3 and any([d.startswith(direction) for d in dirs]), (
            f"The direction argument must be one of {dirs}")
        # degrees
        degrees = math.floor(abs(coord))
        # minutes
        minutes = math.floor((abs(coord) - degrees) * 60)
        # seconds
        seconds = (abs(coord) - degrees - (minutes / 60)) * 3600

        if coord < 0:
            direc = 'S' if direction == 'LAT' else 'W'
        else:
            direc = 'N' if direction == 'LAT' else 'E'

        return degrees, minutes, seconds, direc

    def _degrees_from_hms(self, h: int, m: int, s: float) -> float:
        """
        Find the degrees from the hours, minutes, and seconds of 360 degrees.
        Where as time zones are 15 degrees apart so 24 time zones times 15
        degrees is 360 degrees.

        The angle may be expressed as negative east of the meridian plane
        and positive west of the meridian plane, or as positive westward
        from 0° to 360°. The angle may be measured in degrees or in time,
        with 24h = 360° exactly. So one hour is equal to (360/24)° = 15°.

        :param int h: The hour.
        :param int m: The minute.
        :param float s: The second.
        :returns: The degrees from the hours, minutes, and seconds of
                  360 degrees.
        :rtype: float
        """
        return 15 * h + 15 * m / 60 + 15 * s / 3600

    def _hms_from_degrees(self, deg: float) -> tuple:
        """
        Find the hours, minutes, and seconds from 0 - 360 degrees. Where
        as time zones are 15 degrees apart so 24 time zones times 15
        degrees is 360 degrees.

        The angle may be expressed as negative east of the meridian plane
        and positive west of the meridian plane, or as positive westward
        from 0° to 360°. The angle may be measured in degrees or in time,
        with 24h = 360° exactly. So one hour is equal to (360/24)° = 15°.

        :param float deg: The degrees of the 360 degree curcomference of
                          the earth.
        :returns: The hours, minutes, and seconds.
        :rtype: tuple
        """
        h = math.floor(deg / 15)
        m = math.floor((deg / 15 - h) * 60)
        s = (deg / 15 - h - m / 60) * 3600
        return h, m, s

    def _seconds_from_dhms(self, days: int, hours: int, minutes: int,
                           seconds: float, zone: float=0) -> float:
        """
        Convert days, hours, minutes, and seconds to seconds.

        :param int days: Number of days.
        :param int hours: Number of hours.
        :param int minutes: The number of minutes.
        :param float seconds: The number of seconds with possible fraction.
        :param float zone: The timezone in degrees, defaults to 0 or GMT.
        :returns: The number of seconds.
        :rtype: float
        """
        return (days * 86400+ (hours + zone) * 3600 + minutes * 60
                + seconds + zone * 3600)

    def _dhms_from_seconds(self, seconds: float, zone: float=0) -> tuple:
        """
        Convert seconds into days, hours, minutes, and seconds. Depending
        on the timezone there could be an additional day added.

        :param float seconds: The number of seconds with possible fraction.
        :param float zone: The timezone in degrees, defaults to 0 or GMT.
        :returns: The days, hours, minutes, and seconds.
        :rtype: tuple

        .. note::

           See: https://www.timeanddate.com/time/map/

           Timezones can be from -11 to +14 based on the political timeszones
           as of 2024-08-09.
        """
        seconds += zone * 3600
        hours = math.floor(seconds / 3600)
        m = (seconds - hours * 3600) / 60
        minutes = math.floor(m)

        if hours < 24:
            day = 0
        else:
            hours -= 24
            day = 1
            seconds -= 86400

        return day, hours, minutes, (seconds - hours * 3600) - minutes * 60

    def _tz_decimal_from_dhms(self, days: int, hours: int, minutes: int,
                              seconds: int) -> int:
        """
        Convert days, hours, minutes, and seconds to a decimal number
        representing percentage of one revolution around the Earth. Where
        the number 1 indicates one revolution.

        :param int days: Number of days.
        :param int hours: Number of hours.
        :param int minutes: The number of minutes.
        :param float seconds: The number of seconds with possible fraction.
        :returns: A decimal number.
        :rtype: int

        .. note::

           This method is used in determining time zones.
        """
        return self._seconds_from_dhms(days, hours, minutes, seconds) / 86400

    def _tz_dhms_from_decimal(self, dec: float) -> tuple:
        """
        Convert a decimal number into days, hours, minutes, and seconds
        of a time zone. The decimal number represents the percentage of
        one revolution around the Earth. Where the number 1 indicates
        one revolution.

        :param float dec: A decimal number.
        :returns: The days, hours, minutes, and seconds of a time zone.
        :rtype: tuple

        .. note::

           This method is used in determining time zones.
         """
        return self._dhms_from_seconds(dec * 86400)

    def _hms_from_decimal_day(self, dec: float, *, us: bool=False) -> tuple:
        """
        Convert a decimal day to hours, minutes, and seconds. If this
        method is used for a Julian Period day, 0.5 must be added to the
        value before being passed in.

        :param float dec: A decimal number.
        :param bool us: If False (default) no seperate field for microseconds
                        is returned else return microseconds.
        :returns: A tuple representing the hour, minute, and seconds.
        :rtype: tuple

        .. note::

           If a whole number as in 10.5 is passed in, the value to the left
           of the decimal will be stripped off before calculations are done.
        """
        h = self._PARTIAL_DAY_TO_HOURS(dec)
        hour = math.floor(h)
        m = self._PARTIAL_HOUR_TO_MINUTE(h)
        minute = math.floor(m)
        second = self._PARTIAL_MINUTE_TO_SECOND(m)

        if us:
            microsec = (self._PARTIAL_SECOND_TO_MICROSECOND(second),)
            second = math.floor(second)
        else:
            microsec = ()

        return (hour, minute, second) + microsec

    def _decimal_day_from_hms(self, h: int, m: int, s: float) -> float:
        """
        Convert hours, minutes, and seconds to a decimal day.

        :param ine h: The hour.
        :param int m: The minute.
        :param float s: The second.
        :returns: A decimal value representing the day with a partial that
                  indicates the hours, minutes, and seconds.
        :rtype: float
        """
        return (h * 60 * 60 + m * 60 + s) / 86400

    def _sec_microsec_from_seconds(self, second: float) -> tuple:
        """
        Split the second and microseconds.

        :param float second: The second with a partial indicating the
                             microseconds.
        :returns: The second split between the second and microseconds.
        :rtype: tuple
        """
        p = second % 1
        s = abs(second) - p
        s *= -1 if second < 0 else 1
        us = self._PARTIAL_SECOND_TO_MICROSECOND(second)
        return math.floor(s), us

    def _sin_deg(self, theta: float) -> float:
        """
        Convert a value to sine in degrees.

        :param float thete: The value to convert to degrees.
        :returns: The degrees representing the value provided.
        :rtype: float

        .. rubric:: LISP code from  Reingold &  Dershowitz CC TUE p. 513
        .. code:: lisp

           (defun sin-degrees (theta)
             ;; TYPE angle -> amplitude
             ;; Sine of theta (given in degrees).
             (sin (radians-from-degrees theta)))
        """
        return math.sin(math.radians(theta))

    def _cos_deg(self, theta: float) -> float:
        """
        Convert a value to the cosine in degrees.

        :param float thete: The value to convert to degrees.
        :returns: The degrees representing the value provided.
        :rtype: float

        .. rubric:: LISP code from  Reingold &  Dershowitz CC TUE p. 513
        .. code:: lisp

           (defun cos-degrees (theta)
             ;; TYPE angle -> amplitude
             ;; Cosine of theta (given in degrees).
             (cos (radians-from-degrees theta)))
        """
        return math.cos(math.radians(theta))

    def _sigma(self, lists: tuple, func: object) -> float:
        """
        This gives a summation of a list based on the criteria in the
        provided function.

        :param tuple lists: The list of values to sum.
        :param object func: The function that determins the summation
                            parameters.
        :returns: The summation.
        :rtype: float

        .. rubric:: LISP code from  Reingold &  Dershowitz CC TUE p. 473
        .. code:: lisp

           (defmacro sigma (list body)
             ;; TYPE (list-of-pairs (list-of-reals->real))
             ;; TYPE -> real
             ;; list is of the form ((i1 l1)...(in ln)).
             ;; Sum of body for indices i1...in
             ;; running simultaneously thru lists l1...ln.
             `(apply `+ (mapcar (function (lambda
                                            ,(mapcar `car list)
                                            ,body))
                                ,@(mapcar `cadr list))))
        """
        # Ensure all lists have the same length
        assert len(set(len(lst) for lst in lists)) == 1, (
            "Lists must have the same length")
        return sum(func(*e) for e in zip(*lists))

    def _poly(self, x: float, a: list) -> float:
        """
        This is the Horner method of polynomial used to eliminate the use
        of powers.

        .. note::

           | Instead of:
           | y = A + B * x + C * x^2 + D * x^3 + E * x^4
           | do this
           | y = A + x * (B + x * (C + x * (D + x * E)))

        :param float x: Power of number.
        :param list a: The list of numbers in polynomial.
        :returns: The polynomial result.
        :rtype: float

        .. rubric:: LISP code from  Reingold &  Dershowitz CC TUE p. 473
        .. code:: lisp

           (defun poly (x a)
             ;; TYPE (real list-of-reals) -> real
             ;; Sum powers of x with coefficients (from order 0 up) in list a.
             (if (equal a nil)
                 0
               (+ (first a) (* x (poly x (rest a))))))
        """
        return 0 if not a else a[0] + (x * self._poly(x, a[1:]))

    def _days_in_years(self, y: int, *, alt: bool=False) -> int:
        """
        Find the number of days up to the provided year.

        :param int y: The year to count to.
        :param bool alt: If True use the 4|128 rule else if False use the
                         4|100|400 rule. The default is False.
        :returns: The count of days including year one to the given year.
        :rtype: int

        .. note::

           This method starts the count from year 1 of the Julian Calendar,
           however, it uses one of the two leap rules described above instead
           of the usual Julian Calendar leap year rule of every 4 year.
        """
        n_4 = y // 4

        if alt:
            n_128 = y // 128
            n_leap_years = n_4 - n_128
        else:
            n_100 = y // 100
            n_400 = y // 400
            n_leap_years = n_4 - n_100 + n_400

        a = y - n_leap_years  # Non-leap years
        b = y - a  # Leap years
        return a * 365 + b * 366

    def _meeus_from_exact(self, jd: float) -> int:
        """
        The returned difference value to convert an exact algorithm jd to
        a Meeus algorithm jd. This is added to the exact jd.

        :param float jd: Exact Julian Period day.
        :returns: The difference to subtract from an exact algorithm jd.
        :rtype: int
        """
        jd_diff = (
            (1757641.5, 0), (1794165.5, 1), (1830689.5, 2), (1903738.5, 3),
            (1940262.5, 4), (1976786.5, 5), (2049835.5, 6), (2086359.5, 7),
            (2122883.5, 8), (2195932.5, 9), (2232456.5, 10), (2268980.5, 11),
            (2299158.5, 12),
            )
        diff = 2

        for j, df in jd_diff:
            if jd < j:
                diff = df
                break

        return diff

    def _exact_from_meeus(self, jd: float) -> int:
        """
        The returned difference value to convert a Meeus algorithm jd to
        an exact algorithm jd. This is subtracted from the meeus jd.

        :param float jd: Meeus Julian Period day.
        :returns: The difference to subtract from a Meeus algorithm jd.
        :rtype: int
         """
        jd_diff = (
            (1757642.5, 0), (1794167.5, 1), (1830692.5, 2), (1903742.5, 3),
            (1940267.5, 4), (1976792.5, 5), (2049842.5, 6), (2086367.5, 7),
            (2122892.5, 8), (2195942.5, 9), (2232467.5, 10), (2268992.5, 11),
            (2299160.5, 12),
            )
        diff = 2

        for j, df in jd_diff:
            if jd < j:
                diff = df
                break

        return diff

    def _coterminal_angle(self, value: float) -> float:
        """
        Find the Coterminal Angle from a value that is either more than
        360 or less than 0.

        :param float value: The value that is more than 360 or less than 0.
        :returns: The adjusted angle to be between 0 and 360.
        :rtype: float
        """
        value = math.fmod(value, 360)
        return value + 360 if value < 0 else value

    def _interpolation_from_three(self, y1: float, y2: float, y3: float,
                                  n: float, normalize: bool=False) -> float:
        """
        Interpolate from three terms with a factor.

        :param float y1: 1st of the three parameters.
        :param float y2: 2nd of the three parameters.
        :param float y3: 3rd of the three parameters.
        :param float n: The factor.
        :param bool mormalize: If `False' (default) no normalization is done
                               else if `True` normalize.
        :returns: The three factor interpolation.
        :rtype: float
        """
        a = y2 - y1
        b = y3 - y2

        if normalize:
            a += 360 if a < 0 else 0
            b += 360 if b < 0 else 0

        c = b - a
        return y2 + (n / 2) * (a + b + n * c)

    def _truncate_decimal(self, n: int, places: int) -> int:
        """
        Trucate a decimal to a number of places.

        .. note::

           This is somewhat like rounding, but is used in places where
           rounding gives an invalid results.

        :param int n: Number to truncate.
        :param int places: The number of places to truncate to.
        :returns: The truncated number.
        :rtype: int
        """
        p = int('1' + '0' * places)
        return math.floor(n * p) / p

    def _xor_boolean(self, booleans: tuple) -> bool:
        """
        Test that any number of booleans can all be False or only one True.

        :param tuple booleans: A tuple of booleans that cannot be used
                               together.
        :returns: True if only one of the booleans are True or if None are True
                  else False.
        :rtype: bool
        """
        count = sum(booleans)
        return (count % 2 == 1) or (count == 0)

    def _ordinal_from_jd(self, jd: float, *, _exact: bool=True) -> int:
        """
        Convert a Julian Period day to an ordinal number.

        :param float jd: The Julian Period day.
        :param bool _exact: If True (default) the incoming JD is the more
                            astronomically exact Julin Period day else it
                            is the historically correct (Meeus) JD.
        :returns: The ordinal number relating to the Julian Period day.
        :rtype: int
        """
        jd -= 0 if _exact else self._exact_from_meeus(jd)
        # We add 1 because ordinal date representations satrt at 1 not 0.
        return math.floor(jd - self._JULIAN_CAL_EPOCH) + 1

    def _jd_from_ordinal(self, ordinal: int, *, exact: bool=True) -> float:
        """
        Convert an ordinal number to a Julian Period day.

        :param int ordinal: The ordinal number of days starting with one on the
                            first day od the Julian Calendar.
        :param bool exact: If True (default) the outgoing Julian Period day is
                           the more astronomically exact Julin Period day else
                           if False it is the historically correct (Meeus) JD.
        :returns: The Julian Period day relating to the ordinal number.
        :rtype: float
        """
        # We subtract 1 because ordinal date representations satrt at 1 not 0.
        jd = self._JULIAN_CAL_EPOCH + (ordinal - 1)
        return jd + (0 if exact else self._meeus_from_exact(jd))



# -*- coding: utf-8 -*-
#
# badidatetime/julian_period.py
#
__docformat__ = "restructuredtext en"


class JulianPeriod:
    """
    Implementation of the Julian Period.

    This Julian period is said to be named after Julius Cæsar Scaliger,
    the father of Josephus Justus Scaliger (Born: August 5, 1540, Agen,
    France), who invented the concept. The Julian Period is related to
    the Julian Calendar introduced by Julius Caesar, but it is not the
    same.

    See https://quasar.as.utexas.edu/BillInfo/JulianDatesG.html
        https://core2.gsfc.nasa.gov/time/julian.html
        https://www.tondering.dk/claus/cal/julperiod.php
    """
    # January 1, 2000, at 12:00 TT (Terrestrial Time)
    # TAI = TT - 32.184 seconds
    # January 1, 2000, 11:59:27.816 TAI (International Atomic Time)
    # January 1, 2000, 11:58:55.816 UTC (Coordinated Universal Time)
    # See: https://aa.usno.navy.mil/faq/sun_approx
    _JULIAN_EPOCH = 0  # This is the actual Julian Epoch
    _J2000 = 2451545.0  # TDT
    # 28 (solar cycle) × 19 (lunar cycle) × 15 (indiction cycle) = 7980 years
    _JULIAN_YEAR = 365.25
    _JULIAN_LEAP_YEAR = lambda self, year: year % 4 == 0

    def _julian_centuries(self, jde: float) -> float:
        """
        Calculate the Julian centuries from the Julian day.

        :param float jde: The Julian Period day.
        :returns: The Julian century relative to the day.
        :rtype: float
        """
        return (jde - self._J2000) / 36525

    def _julian_millennia(self, jde: float) -> float:
        """
        Calculate the Julian millennia from the Julian day.

        :param float jde: The Julian Period day.
        :returns: The Julian millennia relative to the day.
        :rtype: float
        """
        return (jde - self._J2000) / 365250
